# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from numbers import Integral
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

import numpy as np
from PIL import Image
from torchvision.datasets.utils import download_and_extract_archive
from typing_extensions import Literal

from ._cifar10_base import _CIFAR10Base
from ._utils import md5_check

__all__ = ["CIFAR10C", "CIFAR100C"]

PathLike = Union[str, Path]
Corruptions = Literal[
    "brightness",
    "contrast",
    "defocus_blur",
    "elastic_transform",
    "fog",
    "frost",
    "gaussian_blur",
    "gaussian_noise",
    "glass_blur",
    "impulse_noise",
    "jpeg_compression",
    "motion_blur",
    "pixelate",
    "saturate",
    "shot_noise",
    "snow",
    "spatter",
    "speckle_noise",
    "zoom_blur",
]

Severity = Literal[1, 2, 3, 4, 5]


class CIFAR10C(_CIFAR10Base):
    """
    A CIFAR-10-based test dataset for benchmarking neural network robustness to common
    corruptions [1]_. For each corruption and for each severity (`1-5`), the
    dataset consists of the 10,000 test images from CIFAR-10 with that corruption
    applied to those images.

    References
    ----------
    .. [1] https://zenodo.org/record/2535967#.YnAe_9rMKbg
    """

    url = "https://zenodo.org/record/2535967/files/CIFAR-10-C.tar"
    filename = "CIFAR-10-C.tar"
    tar_md5: str = "56bf5dcef84df0e2308c6dcbcbbd8499"
    base_folder: str = "CIFAR-10-C"

    _file_to_md5: Dict[str, str] = {
        "brightness": "0a81ef75e0b523c3383219c330a85d48",
        "contrast": "3c8262171c51307f916c30a3308235a8",
        "defocus_blur": "7d1322666342a0702b1957e92f6254bc",
        "elastic_transform": "9421657c6cd452429cf6ce96cc412b5f",
        "fog": "7b397314b5670f825465fbcd1f6e9ccd",
        "frost": "31f6ab3bce1d9934abfb0cc13656f141",
        "gaussian_blur": "c33370155bc9b055fb4a89113d3c559d",
        "gaussian_noise": "ecaf8b9a2399ffeda7680934c33405fd",
        "glass_blur": "7361fb4019269e02dbf6925f083e8629",
        "impulse_noise": "2090e01c83519ec51427e65116af6b1a",
        "jpeg_compression": "2b9cc4c864e0193bb64db8d7728f8187",
        "labels": "c439b113295ed5254878798ffe28fd54",
        "motion_blur": "fffa5f852ff7ad299cfe8a7643f090f4",
        "pixelate": "0f14f7e2db14288304e1de10df16832f",
        "saturate": "1cfae0964219c5102abbb883e538cc56",
        "shot_noise": "3a7239bb118894f013d9bf1984be7f11",
        "snow": "bb238de8555123da9c282dea23bd6e55",
        "spatter": "8a5a3903a7f8f65b59501a6093b4311e",
        "speckle_noise": "ef00b87611792b00df09c0b0237a1e30",
        "zoom_blur": "6ea8e63f1c5cdee1517533840641641b",
    }

    all_corruptions = (
        "brightness",
        "contrast",
        "defocus_blur",
        "elastic_transform",
        "fog",
        "frost",
        "gaussian_blur",
        "gaussian_noise",
        "glass_blur",
        "impulse_noise",
        "jpeg_compression",
        "motion_blur",
        "pixelate",
        "saturate",
        "shot_noise",
        "snow",
        "spatter",
        "speckle_noise",
        "zoom_blur",
    )
    noise_corruptions = (
        "gaussian_noise",
        "impulse_noise",
        "shot_noise",
    )
    blur_corruptions = (
        "defocus_blur",
        "glass_blur",
        "motion_blur",
        "zoom_blur",
    )
    weather_corruptions = (
        "brightness",
        "fog",
        "frost",
        "snow",
    )
    digital_corruptions = (
        "contrast",
        "elastic_transform",
        "jpeg_compression",
        "pixelate",
    )

    # not included in augmix results
    extra_corruptions = (
        "gaussian_blur",
        "saturate",
        "spatter",
        "speckle_noise",
    )

    def __init__(
        self,
        root: PathLike,
        corruption: Corruptions,
        severity: Severity,
        transform: Optional[Callable[[Image.Image], Any]] = None,
        target_transform: Optional[Callable[[int], Any]] = None,
        download: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        root : PathLike
            Root directory of dataset where directory
            ``CIFAR-10-C`` exists or will be saved to if download is set to True.

        corruption : str
            The type of corruption, e.g., "fog". See `CIFAR10C.all_corruptions()` for a full list of corruptions.

        severity : Literal[1, 2, 3, 4, 5]
            The severity-level of the corruption, ranging from 1 to 5.

        transform :  Optional[Callable[[Image], Any]]
            A function/transform that takes in a PIL image
            and returns a transformed version. E.g., ``transforms.RandomCrop``

        target_transform : Optional[Callable]
            A function/transform that takes in a target
            and returns a transformed version.

        download : bool, optional (default=False)
            If true, downloads the dataset from the internet and puts it in root directory.
            If dataset is already downloaded, it is not downloaded again."""

        super().__init__(
            root=str(root), transform=transform, target_transform=target_transform
        )

        self._root = (Path(self.root) / self.base_folder).resolve()
        assert corruption in self._file_to_md5
        self.corruption = corruption

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        if not isinstance(severity, (int, Integral)) and 1 <= severity <= 5:
            raise ValueError(f"`severity` must be an integer in [1, 5]. Got {severity}")

        severity -= 1  # type: ignore

        mmap = np.load(self._root / (corruption + ".npy"), mmap_mode="r")
        size = len(mmap) // 5

        # shape-(N, H, W, C)
        self.data = np.array(mmap[severity * size : (severity + 1) * size])
        self.targets = np.load(self._root / "labels.npy")

    def _check_integrity(self) -> bool:
        img_path = self._root / (self.corruption + ".npy")
        labels_path = self._root / "labels.npy"

        if not img_path.is_file() and not labels_path.is_file():
            return False

        img_md5 = md5_check(img_path)
        label_md5 = md5_check(labels_path)

        return (
            img_md5 == self._file_to_md5[self.corruption]
            and label_md5 == self._file_to_md5["labels"]
        )

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return
        download_and_extract_archive(
            self.url, self.root, filename=self.filename, md5=self.tar_md5
        )

    def extra_repr(self) -> str:
        return "Split: Test"


class CIFAR100C(CIFAR10C):
    """
    A CIFAR-100-based test dataset for benchmarking neural network robustness to common
    corruptions [1]_. For each corruption and for each severity (`1-5`), the
    dataset consists of the 10,000 test images from CIFAR-100 with that corruption
    applied to those images.

    Parameters
    ----------
    root : PathLike
        Root directory of dataset where directory
        ``CIFAR-100-C`` exists or will be saved to if download is set to True.

    corruption : str
        The type of corruption, e.g., "fog". See `CIFAR100C.all_corruptions()` for a full
        list of corruptions.

    severity : Literal[1, 2, 3, 4, 5]
        The severity-level of the corruption, ranging from 1 to 5.

    transform :  Optional[Callable[[Image], Any]]
        A function/transform that takes in an PIL image
        and returns a transformed version. E.g., ``transforms.RandomCrop``

    target_transform : Optional[Callable]
        A function/transform that takes in a target
        and returns a transformed version.

    download: bool, optional (default=False)
        If true, downloads the dataset from the internet and puts it in root directory.
        If dataset is already downloaded, it is not downloaded again.

    References
    ----------
    .. [1] https://zenodo.org/record/2535967#.YnAe_9rMKbg
    """

    url = "https://zenodo.org/record/3555552/files/CIFAR-100-C.tar"
    filename = "CIFAR-100-C.tar"
    tar_md5: str = "11f0ed0f1191edbf9fa23466ae6021d3"
    base_folder: str = "CIFAR-100-C"

    _file_to_md5: Dict[str, str] = {
        "brightness": "f22d7195aecd6abb541e27fca230c171",
        "contrast": "322bb385f1d05154ee197ca16535f71e",
        "defocus_blur": "d923e3d9c585a27f0956e2f2ad832564",
        "elastic_transform": "a0792bd6581f6810878be71acedfc65a",
        "fog": "4efc7ebd5e82b028bdbe13048e3ea564",
        "frost": "3a39c6823bdfaa0bf8b12fe7004b8117",
        "gaussian_blur": "5204ba0d557839772ef5a4196a052c3e",
        "gaussian_noise": "ecc4d366eac432bdf25c024086f5e97d",
        "glass_blur": "0bf384f38e5ccbf8dd479d9059b913e1",
        "impulse_noise": "3b3c210ddfa0b5cb918ff4537a429fef",
        "jpeg_compression": "c851b7f1324e1d2ffddeb76920576d11",
        "labels": "bb4026e9ce52996b95f439544568cdb2",
        "motion_blur": "732a7e2e54152ff97c742d4c388c5516",
        "pixelate": "96c00c60f144539e14cffb02ddbd0640",
        "saturate": "c0697e9fdd646916a61e9c312c77bf6b",
        "shot_noise": "b0a1fa6e1e465a747c1b204b1914048a",
        "snow": "0237be164583af146b7b144e73b43465",
        "spatter": "12ccf41d62564d36e1f6a6ada5022728",
        "speckle_noise": "e3f215b1a0f9fd9fd6f0d1cf94a7ce99",
        "zoom_blur": "0204613400c034a81c4830d5df81cb82",
    }

    classes = (
        "apple",
        "aquarium_fish",
        "baby",
        "bear",
        "beaver",
        "bed",
        "bee",
        "beetle",
        "bicycle",
        "bottle",
        "bowl",
        "boy",
        "bridge",
        "bus",
        "butterfly",
        "camel",
        "can",
        "castle",
        "caterpillar",
        "cattle",
        "chair",
        "chimpanzee",
        "clock",
        "cloud",
        "cockroach",
        "couch",
        "crab",
        "crocodile",
        "cup",
        "dinosaur",
        "dolphin",
        "elephant",
        "flatfish",
        "forest",
        "fox",
        "girl",
        "hamster",
        "house",
        "kangaroo",
        "keyboard",
        "lamp",
        "lawn_mower",
        "leopard",
        "lion",
        "lizard",
        "lobster",
        "man",
        "maple_tree",
        "motorcycle",
        "mountain",
        "mouse",
        "mushroom",
        "oak_tree",
        "orange",
        "orchid",
        "otter",
        "palm_tree",
        "pear",
        "pickup_truck",
        "pine_tree",
        "plain",
        "plate",
        "poppy",
        "porcupine",
        "possum",
        "rabbit",
        "raccoon",
        "ray",
        "road",
        "rocket",
        "rose",
        "sea",
        "seal",
        "shark",
        "shrew",
        "skunk",
        "skyscraper",
        "snail",
        "snake",
        "spider",
        "squirrel",
        "streetcar",
        "sunflower",
        "sweet_pepper",
        "table",
        "tank",
        "telephone",
        "television",
        "tiger",
        "tractor",
        "train",
        "trout",
        "tulip",
        "turtle",
        "wardrobe",
        "whale",
        "willow_tree",
        "wolf",
        "woman",
        "worm",
    )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Callable, Iterable, List, Mapping, TypeVar, Union

import torch as tr

T = TypeVar("T", bound=Callable)
NoneType = type(None)


def flatten_params(
    iter_params_or_groups: Union[
        tr.Tensor, Iterable[tr.Tensor], Iterable[Mapping[str, Iterable[tr.Tensor]]]
    ]
) -> List[tr.Tensor]:

    if isinstance(iter_params_or_groups, tr.Tensor):
        return [iter_params_or_groups]

    out = []
    for params in iter_params_or_groups:
        if isinstance(params, tr.Tensor):
            out.append(params)
        else:
            out.extend(params["params"])
    return out



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from dataclasses import is_dataclass
from inspect import Parameter, signature
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

from hydra._internal.callbacks import Callbacks
from hydra._internal.config_loader_impl import ConfigLoaderImpl
from hydra._internal.hydra import Hydra
from hydra._internal.utils import create_config_search_path
from hydra.core.config_store import ConfigStore
from hydra.core.utils import JobReturn
from hydra.plugins.sweeper import Sweeper
from hydra.types import HydraContext, RunMode
from hydra_zen import instantiate
from hydra_zen.typing._implementations import DataClass_
from omegaconf import OmegaConf
from typing_extensions import Literal


def is_config(cfg: Any) -> bool:
    return is_dataclass(cfg) or OmegaConf.is_config(cfg)


@overload
def launch(
    config: Union[DataClass_, Type[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[List[str]] = None,
    config_name: str = "zen_launch",
    job_name: str = "zen_launch",
    with_log_configuration: bool = True,
    multirun: Literal[False] = False,
) -> JobReturn:
    ...


@overload
def launch(
    config: Union[DataClass_, Type[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[List[str]] = None,
    config_name: str = "zen_launch",
    job_name: str = "zen_launch",
    with_log_configuration: bool = True,
    multirun: Literal[True] = True,
) -> List[List[JobReturn]]:
    ...


def launch(
    config: Union[DataClass_, Type[DataClass_], Mapping[str, Any]],
    task_function: Callable[[Any], Any],
    overrides: Optional[List[str]] = None,
    config_name: str = "zen_launch",
    job_name: str = "zen_launch",
    with_log_configuration: bool = True,
    multirun: bool = False,
) -> Union[JobReturn, List[List[JobReturn]]]:

    cs = ConfigStore().instance()
    cs.store(name=config_name, node=config)
    search_path = create_config_search_path(None)

    # initiate Hydra without enforcing single instance
    config_loader = ConfigLoaderImpl(config_search_path=search_path)
    hydra = Hydra(task_name=job_name, config_loader=config_loader)

    if not multirun:
        # Here we can use Hydra's `run` method
        job = hydra.run(
            config_name=config_name,
            task_function=task_function,
            overrides=overrides if overrides is not None else [],
            with_log_configuration=with_log_configuration,
        )

    else:
        # Instead of running Hydra's `multirun` method we instantiate
        # and run the sweeper method.  This allows us to run local
        # sweepers and launchers without installing them in `hydra_plugins`
        # package directory.
        cfg = hydra.compose_config(
            config_name=config_name,
            overrides=overrides if overrides is not None else [],
            with_log_configuration=with_log_configuration,
            run_mode=RunMode.MULTIRUN,
        )

        callbacks = Callbacks(cfg)
        callbacks.on_multirun_start(config=cfg, config_name=config_name)

        # Instantiate sweeper without using Hydra's Plugin discovery (Zen!)
        sweeper = instantiate(cfg.hydra.sweeper)
        assert isinstance(sweeper, Sweeper)
        sweeper.setup(
            config=cfg,
            hydra_context=HydraContext(
                config_loader=hydra.config_loader, callbacks=callbacks
            ),
            task_function=task_function,
        )

        task_overrides = OmegaConf.to_container(cfg.hydra.overrides.task, resolve=False)
        assert isinstance(task_overrides, list)

        job = sweeper.sweep(arguments=task_overrides)
        callbacks.on_multirun_end(config=cfg, config_name=config_name)

    return job


T1 = TypeVar("T1")


# TODO: - zen's instantiation should be memoized so that subsequent access
#         to an attribute returns the same instance
#       - zen should interface with a singelton that "recognizes"
#         the configs that it interacts with
class zen(Generic[T1]):
    def __init__(self, func: Callable[..., T1]) -> None:
        self.func = func
        self.parameters = signature(self.func).parameters

    def validate(self, cfg: Any, excluded_params: Iterable[str] = ()):
        excluded_params = set(excluded_params)

        if callable(cfg):
            cfg = cfg()  # instantiate dataclass to result default-factory

        num_pos_only = sum(
            p.kind is p.POSITIONAL_ONLY for p in self.parameters.values()
        )
        if num_pos_only:
            assert hasattr(cfg, "_args_")
            assert len(cfg._args_) == num_pos_only

        missing_params: List[str] = []
        for name, param in self.parameters.items():
            if name in excluded_params:
                continue

            if param.kind is Parameter.POSITIONAL_ONLY:
                continue
            if not hasattr(cfg, name) and param.default is param.empty:
                missing_params.append(name)

        if missing_params:
            raise TypeError(
                f"`cfg` is missing the following fields: {', '.join(missing_params)}"
            )

    def __call__(self, cfg: Any, *args: Any, **kwargs: Any) -> T1:
        if callable(cfg):
            cfg = cfg()  # instantiate dataclass to resolve default-factory

        if not args:
            args_ = getattr(cfg, "_args_", [])
            assert isinstance(args_, Sequence)
            args_ = list(args_)
        else:
            args_ = list(args)

        del args

        cfg_kwargs = {
            name: (
                getattr(cfg, name, param.default)
                if param.default is not param.empty
                else getattr(cfg, name)
            )
            for name, param in self.parameters.items()
            if param.kind is not param.POSITIONAL_ONLY and name not in kwargs
        }

        # instantiate any configs
        args_ = [instantiate(x) if is_config(x) else x for x in args_]
        cfg_kwargs = {
            name: instantiate(val) if is_config(val) else val
            for name, val in cfg_kwargs.items()
        }
        kwargs = {
            name: instantiate(val) if is_config(val) else val
            for name, val in kwargs.items()
        }
        return self.func(*args_, **cfg_kwargs, **kwargs)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from itertools import product
from typing import Iterable, Iterator, NamedTuple, Optional, Tuple

import numpy as np
from numpy.typing import DTypeLike


class FourierBasis(NamedTuple):
    """
    Describes a 2D planewave generated via:

    A * cos(2pi * freq_vector @ x + phase_shift)

    Where `A` normalizes the 2D array to have an :math:`L^2`-norm of 1,
    and `x` is evaluated on a grid of positions on [0, H] x [0, W].

    Parameters
    ----------
    basis: np.ndarray, shape-(H, W)
        The amplitudes of the 2D planewave.

    position : Tuple[int, int]
        The (k-row, k-col) coordinate in shifted k-space, where
        the freq-0 component has been shifted to reside at the
        center of the shape-(H, W) "k-image".

    sym_position : Tuple[int, int]
        The k-space coordinate symmetric to `position`, about the
        center of the k-space image.

    phase_shift : float
        The total phase-shift applied to the cosine-generated
        planewave.
    """

    basis: np.ndarray
    position: Tuple[int, int]
    sym_position: Tuple[int, int]
    phase_shift: float


def generate_fourier_bases(
    nrows: int,
    ncols: int,
    dtype: DTypeLike = np.float32,
    *,
    factor_2pi_phase_shift: float = 0,
    row_col_coords: Optional[Iterable[Tuple[int, int]]] = None,
) -> Iterator[FourierBasis]:
    """Yields each unique, real-valued 2D array with unit norm, such that its Fourier
    transform is supported at each (i, j) and its symmetric position on a shape-(nrows, ncols)
    grid in Fourier space, where the lowest frequency component resides at the center of
    the grid.

    Parameters
    ----------
    nrows : int

    ncols : int

    dtype : DTypeLike

    factor_2pi_phase_shift : float (optional, default=0)
        E.g., specifying ``0.5`` will induce a phase-shift of :math:`\pi` on all bases.

    row_col_coords : Optional[Iterable[Tuple[int, int]]]
        If provided, specifies the specific Fourier-bases to generate in terms of their
        Fourier-space locations. Otherwise, all non-redundant entries on the nrows x ncols
        grip will be generated.

    Yields
    ------
    fourier_basis: FourierBasis
        Yields the array's position in Fourier space, (k_row, k_col), its symmetric position, and the
        shape-(nrows, ncols) real-valued basis array itself.

    Notes
    -----
    The iteration is performed in row-major order in 2D Fourier space, starting with the position (0, 0).

    Examples
    --------
    Here we will visualize the collection of bases in Fourier space that support the

    >>> import numpy as np
    >>> from rai_toolbox.augmentations.fourier import generate_fourier_bases
    >>> [np.round(out.basis, 2) for out in generate_fourier_bases(3, 3)]
    [array([[ 0.33,  0.12, -0.46],
        [ 0.12, -0.46,  0.33],
        [-0.46,  0.33,  0.12]]),
    array([[ 0.33,  0.33,  0.33],
            [ 0.12,  0.12,  0.12],
            [-0.46, -0.46, -0.46]]),
    array([[ 0.33, -0.46,  0.12],
            [ 0.12,  0.33, -0.46],
            [-0.46,  0.12,  0.33]]),
    array([[ 0.33,  0.12, -0.46],
            [ 0.33,  0.12, -0.46],
            [ 0.33,  0.12, -0.46]]),
    array([[0.33, 0.33, 0.33],
            [0.33, 0.33, 0.33],
            [0.33, 0.33, 0.33]])]

    We'll now inspect the corresponding support arrays in Fourier space.
    Note that only five basis arrays, not nine, are returned in association with the
    specified 3x3 grid. This is because the remaining four bases are redundant due
    to the symmetries in the support space.

    >>> f = lambda x: np.round(np.abs(np.fft.fftshift(np.fft.fft2(x))))
    >>> [f(out.basis) for out in generate_fourier_bases(3, 3)]
    [array([[2., 0., 0.],
            [0., 0., 0.],
            [0., 0., 2.]]),
    array([[0., 2., 0.],
            [0., 0., 0.],
            [0., 2., 0.]]),
    array([[0., 0., 2.],
            [0., 0., 0.],
            [2., 0., 0.]]),
    array([[0., 0., 0.],
            [2., 0., 2.],
            [0., 0., 0.]]),
    array([[0., 0., 0.],
            [0., 3., 0.],
            [0., 0., 0.]])]
    """
    pos = np.array(list(np.ndindex(nrows, ncols))).reshape(nrows, ncols, 2)
    shape = np.array((nrows, ncols))

    marker = np.zeros([nrows, ncols], dtype=np.uint8)

    if row_col_coords is None:
        row_col_coords = product(range(nrows), range(ncols))

    for i, j in row_col_coords:
        if marker[i, j] > 0:
            continue

        k = np.array([i, j])

        # Given a point, (kx, ky), in 2D Fourier space for real-valued data, compute
        # the symmetric point (kx', ky'). I.e. the Fourier coefficients at these points
        # are related by complex conjugation: c_{kx, ky} = c^*_{kx', ky'}
        sym_i, sym_j = np.where(
            np.mod(shape, 2) == 0, np.mod(shape - k, shape), shape - 1 - k
        )

        marker[i, j] = 1
        marker[sym_i, sym_j] = 1

        phase_shift = 1 / 4 * np.pi + (2 * np.pi * factor_2pi_phase_shift)

        # shift k so 0-freq resides at k=(Lx/2, Ly/2)
        k = k + np.ceil(shape / 2)

        # basis = A * np.cos((2 * np.pi) * (pos @ (k / shape)) + phase_shift)
        basis = pos @ (k / shape)
        basis *= 2 * np.pi
        basis += phase_shift
        np.cos(basis, out=basis)

        basis /= np.linalg.norm(basis, keepdims=True)

        # TODO: calculate frequency-vector to make it simple for users
        # to deduce per-image-length frequency of each basis
        #
        # This works for even-valued image-shapes:
        #
        # freq_vector=tuple(
        #     float(ele) if ele <= np.ceil(s / 2) else abs(s - ele)
        #     for ele, s in zip(k, shape)
        # )
        #

        yield FourierBasis(
            basis=basis.astype(dtype),
            position=(i, j),
            sym_position=(sym_i, sym_j),
            phase_shift=phase_shift,
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Any, Callable, Optional, Union

import numpy as np
from PIL import Image
from torchvision.datasets.utils import download_url

from ._cifar10_base import _CIFAR10Base
from ._utils import md5_check

PathLike = Union[str, Path]

__all__ = ["CIFAR10P1"]


class CIFAR10P1(_CIFAR10Base):
    """CIFAR-10.1 dataset is a new test set for CIFAR-10.

    "CIFAR-10.1 contains 2,000 new test images that were sampled after multiple years of
    research on the original CIFAR-10 dataset. The data collection for CIFAR-10.1 was
    designed to minimize distribution shift relative to the original dataset."

    See https://github.com/modestyachts/CIFAR-10.1 for more details.
    """

    _data_md5 = "4fcae82cb1326aec9ed1dc1fc62345b8"
    _labels_md5 = "09a97fb7c430502fcbd69d95093a3f85"
    url = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/"
    _data_filename = "cifar10.1_v6_data.npy"
    _labels_filename = "cifar10.1_v6_labels.npy"

    def __init__(
        self,
        root: PathLike,
        transform: Optional[Callable[[Image.Image], Any]] = None,
        target_transform: Optional[Callable[[int], Any]] = None,
        download: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        root : PathLike
            Root directory of dataset where
            CIFAR-10.1 files exist or will be saved to if download is set to True.

        transform :  Optional[Callable[[Image], Any]]
            A function/transform that takes in a PIL image
            and returns a transformed version. E.g., ``transforms.RandomCrop``

        target_transform : Optional[Callable]
            A function/transform that takes in a target
            and returns a transformed version.

        download : bool, optional (default=False)
            If true, downloads the dataset from the internet and puts it in root directory.
            If dataset is already downloaded, it is not downloaded again."""

        super().__init__(
            root=str(root), transform=transform, target_transform=target_transform
        )

        self._root = Path(self.root).resolve()
        self._img_path = self._root / self._data_filename
        self._labels_path = self._root / self._labels_filename

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError(
                "Dataset not found or corrupted."
                + " You can use download=True to download it"
            )

        # shape-(N, H, W, C)
        self.data = np.load(self._img_path)
        self.targets = np.load(self._labels_path)

    def _check_integrity(self) -> bool:
        img_path = self._root / self._data_filename
        labels_path = self._root / self._labels_filename

        if not img_path.is_file() and not labels_path.is_file():
            return False

        img_md5 = md5_check(img_path)
        label_md5 = md5_check(labels_path)

        return img_md5 == self._data_md5 and label_md5 == self._labels_md5

    def download(self) -> None:
        if self._check_integrity():
            print("Files already downloaded and verified")
            return
        download_url(
            self.url + self._data_filename,
            self.root,
            filename=self._data_filename,
            md5=self._data_md5,
        )
        download_url(
            self.url + self._labels_filename,
            self.root,
            filename=self._labels_filename,
            md5=self._labels_md5,
        )

    def extra_repr(self) -> str:
        return "Split: Test"



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from functools import partial
from typing import Any, Dict, Optional, Union

import torch
from torch import Generator, Tensor, default_generator
from torch.optim import SGD
from typing_extensions import Final

from rai_toolbox._typing import Optimizer as Opt
from rai_toolbox._typing import OptimizerType, OptimParams, Partial
from rai_toolbox._utils import check_param_group_value, value_check

from .misc import TopQGradientOptimizer
from .optimizer import (
    REQUIRED,
    ChainedParamTransformingOptimizer,
    DatumParamGroup,
    ParamTransformingOptimizer,
)

__all__ = [
    "L1NormedGradientOptim",
    "L2NormedGradientOptim",
    "SignedGradientOptim",
    "L2ProjectedOptim",
    "LinfProjectedOptim",
    "L1NormedGradientOptim",
    "L1qNormedGradientOptim",
]
_TINY = torch.finfo(torch.float32).tiny


class _HasEpsilon(DatumParamGroup):
    epsilon: float


class _LpNormOptimizer(ParamTransformingOptimizer):
    r"""A base optimizer whose step normalizes based on the p-norm:

    .. math::

       ||x||_p = (|x|_1^{p}, ...,  |x|_n^{p})^{1/p}

    When performing gradient-based updates to the optimizer's parameters,
    the gradients are normalized by the p-norm.
    """

    _p: Union[int, float]

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        param_ndim: Optional[int] = -1,
        defaults: Optional[Dict[str, Any]] = None,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        div_by_zero_eps: float = _TINY,
        **kwargs,
    ):
        """
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        param_ndim : Optional[int]
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            A lower bound used to clamp the normalization factor to prevent div-by-zero.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.
        """
        if not hasattr(self, "_p"):
            raise TypeError(f"{type(self).__name__} must have the attribute `_p` set.")
        else:
            if not isinstance(self.p, (int, float)):
                raise TypeError(
                    f"{type(self).__name__}.p must be an int or float, got {self.p}"
                )

        super().__init__(
            params,
            InnerOpt=InnerOpt,
            param_ndim=param_ndim,
            defaults=defaults,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            **kwargs,
        )

        self.div_by_zero_eps = value_check("div_by_zero_eps", div_by_zero_eps, min_=0.0)

    @property
    def p(self) -> Union[float, int]:
        return self._p

    def per_datum_norm(self, x: torch.Tensor) -> torch.Tensor:
        return torch.norm(x, p=self.p, dim=1)  # type: ignore

    def _pre_step_transform_(self, param: Tensor, **_unused_kwargs) -> None:
        if param.grad is None:  # pragma: no cover
            return

        g = param.grad.flatten(1)
        g_norm = self.per_datum_norm(g).view(-1, *([1] * (param.ndim - 1)))
        param.grad /= torch.clamp(g_norm, self.div_by_zero_eps, None)


class SignedGradientOptim(ParamTransformingOptimizer):
    r"""A gradient-tranforming optimizer that takes the elementwise sign
    of a parameter's gradient prior to using `InnerOp.step` to update the
    corresponding parameter.

    See Also
    --------
    L1NormedGradientOptim
    L2NormedGradientOptim
    ParamTransformingOptimizer
    """

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        defaults: Optional[Dict[str, Any]] = None,
        param_ndim: Optional[int] = None,
        **inner_opt_kwargs,
    ) -> None:
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        param_ndim : Optional[int]
            Controls how `_pre_step_transform_` is broadcast onto the gradient
            of a given parameter. This has no effect for `SignedGradientOptim`.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        Let's create use `SignedGradientOptim` along with a SGD-step with a
        learning rate of `1.0`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import SignedGradientOptim

        Creating a parameter for our optimizer to update, and our optimizer.

        >>> x = tr.tensor([-1.5, 1.5], requires_grad=True)
        >>> optim = SignedGradientOptim([x], InnerOpt=tr.optim.SGD, lr=1.0)

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([-2.0, 20.0]) * x).sum().backward()
        >>> x.grad # the original gradient
        tensor([-2., 20.])

        Performing a step with our optimizer transforms the gradient in-place, and then updates the parameter using `SGD([x], lr=1.0).step()`.

        >>> optim.step()
        >>> x.grad # the normalized gradient
        tensor([-1.,  1.])
        >>> x  # the updated parameter
        tensor([-0.5000,  0.5000], requires_grad=True)
        """
        super().__init__(
            params,
            InnerOpt,
            param_ndim=param_ndim,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            defaults=defaults,
            **inner_opt_kwargs,
        )

    def _pre_step_transform_(self, param: Tensor, **_unused_kwargs) -> None:
        if param.grad is None:  # pragma: no cover
            return

        torch.sign(param.grad, out=param.grad)


class L1NormedGradientOptim(_LpNormOptimizer):
    r"""A gradient-tranforming optimizer that normalizes the gradient by
    its :math:`L^1`-norm prior to using `InnerOp.step` to update the
    corresponding parameter.

    See Also
    --------
    L2NormedGradientOptim
    SignedGradientOptim
    ParamTransformingOptimizer

    Examples
    --------
    Let's create an optimizer that normalizes all parameter gradients using
    their :math:`L^1`-norm, and then updates the parameters with a standard
    SGD-step with a learning rate of `1.0`.

    >>> import torch as tr
    >>> from rai_toolbox.optim import L1NormedGradientOptim

    Creating a parameter for our optimizer to update, and our optimizer. We
    want the norm to be computed over the entire gradient tensor – without
    broadcasting – so we specify `param_ndim=None`.

    >>> x = tr.tensor([-1.0, 1.0], requires_grad=True)
    >>> optim = L1NormedGradientOptim([x], param_ndim=None, InnerOpt=tr.optim.SGD, lr=1.0)

    Performing a simple calculation with `x` and performing backprop to create
    a gradient.

    >>> (tr.tensor([2.0, 2.0]) * x).sum().backward()
    >>> x.grad # the un-normed gradient
    tensor([2., 2.])

    Performing a step with our optimizer transforms the gradient in-place, and then updates the parameter using `SGD([x], lr=1.0).step()`.

    >>> optim.step()
    >>> x.grad # the normalized gradient
    tensor([0.5000, 0.5000])
    >>> x  # the updated parameter
    tensor([-1.5000,  0.5000], requires_grad=True)
    """
    _p: Final = 1


class L2NormedGradientOptim(_LpNormOptimizer):
    r"""A gradient-tranforming optimizer that normalizes the gradient by
    its :math:`L^2`-norm prior to using `InnerOp.step` to update the
    corresponding parameter.

    The transformation is applied to the gradient in accordance with `param_ndim`.

    See Also
    --------
    L1NormedGradientOptim
    SignedGradientOptim
    ParamTransformingOptimizer

    Examples
    --------
    Let's create an optimizer that normalizes all parameter gradients using
    their :math:`L^2`-norm, and then updates the parameters with a standard
    SGD-step with a learning rate of `1.0`.

    >>> import torch as tr
    >>> from rai_toolbox.optim import L2NormedGradientOptim

    Creating a parameter for our optimizer to update, and our optimizer. We
    want the norm to be computed over the entire gradient tensor – without
    broadcasting – so we specify `param_ndim=None`.

    >>> x = tr.tensor([-1.0, 1.0], requires_grad=True)
    >>> optim = L2NormedGradientOptim([x], param_ndim=None, InnerOpt=tr.optim.SGD, lr=1.0)

    Performing a simple calculation with `x` and performing backprop to create
    a gradient.

    >>> (tr.tensor([2.0, 2.0]) * x).sum().backward()
    >>> x.grad # the un-normed gradient
    tensor([2., 2.])

    Performing a step with our optimizer transforms the gradient in-place, and then updates the parameter using `SGD([x], lr=1.0).step()`.

    >>> optim.step()
    >>> x.grad # the normalized gradient
    tensor([0.7071, 0.7071])
    >>> x  # the updated parameter
    tensor([-1.7071,  0.2929], requires_grad=True)
    """
    _p: Final = 2


class L2ProjectedOptim(L2NormedGradientOptim):
    r"""A gradient-tranforming optimizer that constrains the updated parameters
    to lie within an :math:`\epsilon`-sized ball in :math:`L^2` space centered on the
    origin.

    A step with this optimizer normalizes the gradient by its :math:`L^2`-norm
    prior to using `InnerOp.step` to update the corresponding parameter. Each parameter
    is then projected into the constraint set.

    The transformation/projection is applied to the gradient/parameter in accordance
    with `param_ndim`.

    See Also
    --------
    L2NormedGradientOptim
    LinfProjectedOptim
    ParamTransformingOptimizer
    """

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        epsilon: float = REQUIRED,
        param_ndim: Union[int, None] = -1,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        defaults: Optional[Dict[str, Any]] = None,
        div_by_zero_eps: float = _TINY,
        **inner_opt_kwargs,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        epsilon : float
            Specifies the size of the L2-space ball that all parameters will be
            projected into, post optimization step.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            A lower bound used to clamp the normalization factor to prevent div-by-zero.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        Let's create an optimizer that normalizes all parameter gradients using
        their :math:`L^2`-norm, and then updates the parameters with a standard
        SGD-step with a learning rate of `1.0`. After the step, each parameter
        will be projected into a :math:`L^2`-ball of radius `0.8`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L2ProjectedOptim

        Creating a parameter for our optimizer to update, and our optimizer. We
        want the norm to be computed over the entire gradient tensor – without
        broadcasting – so we specify `param_ndim=None`. This also controls the
        projection behavior.

        >>> x = tr.tensor([-1.0, 1.0], requires_grad=True)
        >>> optim = L2ProjectedOptim([x], param_ndim=None, InnerOpt=tr.optim.SGD, lr=1.0, epsilon=0.8)

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([2.0, 2.0]) * x).sum().backward()
        >>> x.grad # the un-normed gradient
        tensor([2., 2.])

        Performing a step with our optimizer transforms the gradient in-place, updates the
        parameter using `SGD([x], lr=1.0).step()`, and then projects the parameter into
        the constraint set.

        >>> optim.step()
        >>> x.grad # the normalized gradient
        tensor([0.7071, 0.7071])
        >>> x  # the updated parameter
        tensor([-0.7885,  0.1353], requires_grad=True)
        >>> x.norm(p=2).item() # `x` lies on the L2-ball of radius 0.8
        0.800000011920929
        """

        if defaults is None:
            defaults = {}
        defaults.setdefault("epsilon", epsilon)

        super().__init__(
            params,
            InnerOpt=InnerOpt,
            defaults=defaults,
            param_ndim=param_ndim,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            div_by_zero_eps=div_by_zero_eps,
            **inner_opt_kwargs,
        )
        check_param_group_value("epsilon", self.param_groups, min_=0.0)

    def _post_step_transform_(self, param: Tensor, optim_group: _HasEpsilon) -> None:
        """Applies an in-place projection on the given parameter"""
        param.renorm_(p=self.p, dim=0, maxnorm=optim_group["epsilon"])


class LinfProjectedOptim(SignedGradientOptim):
    r"""A gradient-tranforming optimizer that constrains the updated parameter values to fall within :math:`[-\epsilon, \epsilon]`.

    A step with this optimizer takes the elementwise sign of a parameter's gradient
    prior to using `InnerOp.step` to update the corresponding parameter. The updated
    parameter is then clamped elementwise to :math:`[-\epsilon, \epsilon]`.

    See Also
    --------
    L2NormedGradientOptim
    LinfProjectedOptim
    ParamTransformingOptimizer
    """

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        epsilon: float = REQUIRED,
        param_ndim=None,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        defaults: Optional[Dict[str, Any]] = None,
        **inner_opt_kwargs,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        epsilon : float
            Specifies the size of the L2-space ball that all parameters will be
            projected into, post optimization step.

        param_ndim : Optional[int]
            Clamp is performed elementwise, and thus `param_ndim` need not be adjusted.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        Let's use `LinfProjectedOptim` along with a standard SGD-step with a learning rate
        of `1.0`. After the step, each parameter will have its values clamped to :math:`[-1.8, 1.8]`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L2ProjectedOptim

        Creating a parameter for our optimizer to update, and our optimizer. We
        specify `epsilon=1.8` so that the parameters are projected to the desired domain.

        >>> x = tr.tensor([-1.0, 0.5], requires_grad=True)
        >>> optim = LinfProjectedOptim([x], epsilon=1.8, InnerOpt=tr.optim.SGD, lr=1.0)

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> (tr.tensor([2.0, -2.0]) * x).sum().backward()
        >>> x.grad # the un-normed gradient
        tensor([2., -2.])

        Performing a step with our optimizer transforms the gradient in-place, updates the
        parameter using `SGD([x], lr=1.0).step()`, and then projects the parameter into
        the constraint set.

        >>> optim.step()
        >>> x.grad # the normalized gradient
        tensor([1.0, -1.0])
        >>> x  # the updated parameter
        tensor([-1.8000,  1.5000], requires_grad=True)
        """
        if defaults is None:
            defaults = {}
        defaults.setdefault("epsilon", epsilon)

        super().__init__(
            params,
            InnerOpt=InnerOpt,
            defaults=defaults,
            param_ndim=param_ndim,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            **inner_opt_kwargs,
        )

        check_param_group_value("epsilon", self.param_groups, min_=0.0)

    def _post_step_transform_(self, param: Tensor, optim_group: _HasEpsilon) -> None:
        epsilon = optim_group["epsilon"]
        param.clamp_(min=-epsilon, max=epsilon)


class L1qNormedGradientOptim(ChainedParamTransformingOptimizer):
    r"""A gradient-transforming optimizer that sparsifies a parameter's gradient and
    normalizes the gradient to have an :math:`L^1`-norm of `grad_scale`, prior to
    updating the parameter using `InnerOpt.step`.

    The sparsification process retains only the signs (i.e., :math:`\pm 1`) of the
    gradient's elements. The transformation is applied to the gradient in accordance
    with `param_ndim`.

    See Also
    --------
    L1NormedGradientOptim
    L2NormedGradientOptim
    TopQGradientOptimizer
    ParamTransformingOptimizer
    """

    def __init__(
        self,
        params: OptimParams,
        InnerOpt: Union[Partial[Opt], OptimizerType] = SGD,
        *,
        q: float = REQUIRED,
        dq: float = 0.0,
        param_ndim: Union[int, None] = -1,
        grad_scale: float = 1.0,
        grad_bias: float = 0.0,
        defaults: Optional[Dict[str, Any]] = None,
        div_by_zero_eps: float = _TINY,
        generator: Generator = default_generator,
        **inner_opt_kwargs,
    ):
        r"""
        Parameters
        ----------
        params : Sequence[Tensor] | Iterable[Mapping[str, Any]]
            Iterable of parameters or dicts defining parameter groups.

        InnerOpt : Type[Optimizer] | Partial[Optimizer], optional (default=`torch.nn.optim.SGD`)
            The optimizer that updates the parameters after their gradients have
            been transformed.

        q : float
            Specifies the (fractional) percentile of absolute-largest gradient elements
            to retain when sparsifying the gradient. E.g., `q=0.9` means that only the
            gradient elements within the 90th-percentile will be retained.

            Must be within `[0.0, 1.0]`. The sparsification is applied to the gradient
            in accordance to `param_ndim`.

        dq : float, optional (default=0.0)
            If specified, the sparsity factor for each gradient transformation will
            be drawn from a uniform distribution over :math:`[q - dq, q + dq] \in [0.0, 1.0]`.

        param_ndim : Union[int, None], optional (default=-1)
            Determines how a parameter and its gradient is temporarily reshaped prior
            to being passed to both `_pre_step_transform_` and `_post_step_transform_`.
            By default,the transformation broadcasts over the tensor's first dimension
            in a batch-like style. This can be specified per param-group

            - A positive number determines the dimensionality of the tensor that the transformation will act on.
            - A negative number indicates the 'offset' from the dimensionality of the tensor (see "Notes" for examples).
            - `None` means that the transformation will be applied directly to the tensor without any broadcasting.

            See `ParamTransformingOptimizer` for more details and examples.

        grad_scale : float, optional (default=1.0)
            Multiplies each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        grad_bias : float, optional (default=0.0)
            Added to each gradient in-place after the in-place transformation is
            performed. This can be specified per param-group.

        defaults : Optional[Dict[str, Any]]
            Specifies default parameters for all parameter groups.

        div_by_zero_eps : float, optional (default=`torch.finfo(torch.float32).tiny`)
            A lower bound used to clamp the normalization factor to prevent div-by-zero.

        generator : torch.Generator, optional (default=`torch.default_generator`)
            Controls the RNG source.

        **inner_opt_kwargs : Any
            Named arguments used to initialize `InnerOpt`.

        Examples
        --------
        Let's use `L1qNormedGradientOptim` along with a standard SGD-step with a learning
        rate of `1.0`. We'll sparsify the gradient to retain the top 70% elements of the
        tensor, and we'll normalize the sparse gradient to have a :math:`L^1`-norm of `1.8`.

        >>> import torch as tr
        >>> from rai_toolbox.optim import L1qNormedGradientOptim

        Creating a parameter for our optimizer to update, and our optimizer. We
        specify `param_ndim=None` so that the sparsification/normalization occurs on the
        gradient without any broadcasting.

        >>> x = tr.tensor([1.0, 1.0, 1.0], requires_grad=True)
        >>> optim = L1qNormedGradientOptim(
        ...     [x],
        ...     q=0.30,
        ...     grad_scale=1.8,
        ...     InnerOpt=tr.optim.SGD,
        ...     lr=1.0,
        ...     param_ndim=None,
        ... )

        Performing a simple calculation with `x` and performing backprop to create
        a gradient.

        >>> x.backward(gradient=tr.tensor([0.0, 1.0, 2.0]))
        >>> x.grad # the original gradient
        tensor([0., 1., 2.])

        Performing a step with our optimizer sparsifies and normalizes the gradient in-place, and then updates the parameter using `SGD([x], lr=1.0).step()`.

        >>> optim.step()
        >>> x.grad # the signed, sparsified, and normalized gradient
        tensor([0.0000, 0.9000, 0.9000])
        >>> x  # the updated parameter
        tensor([1.0000, 0.1000, 0.1000], requires_grad=True)
        """
        super().__init__(
            partial(TopQGradientOptimizer, q=q, dq=dq, generator=generator),
            SignedGradientOptim,
            partial(L1NormedGradientOptim, div_by_zero_eps=div_by_zero_eps),
            params=params,
            InnerOpt=InnerOpt,
            defaults=defaults,
            param_ndim=param_ndim,
            grad_scale=grad_scale,
            grad_bias=grad_bias,
            **inner_opt_kwargs,
        )



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from .models import AdditivePerturbation, PerturbationModel
from .init import (
    uniform_like_l1_n_ball_,
    uniform_like_l2_n_ball_,
    uniform_like_linf_n_ball_,
)
from .solvers import gradient_ascent, random_restart

__all__ = [
    "AdditivePerturbation",
    "PerturbationModel",
    "gradient_ascent",
    "random_restart",
    "uniform_like_l1_n_ball_",
    "uniform_like_l2_n_ball_",
    "uniform_like_linf_n_ball_",
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Any, Dict, Tuple

import numpy as np
from PIL import Image
from torchvision.datasets.vision import VisionDataset

__all__ = ["_CIFAR10Base"]


class _CIFAR10Base(VisionDataset):
    classes = (
        "airplane",
        "automobile",
        "bird",
        "cat",
        "deer",
        "dog",
        "frog",
        "horse",
        "ship",
        "truck",
    )
    data: np.ndarray  # shape-(N, H, W, C)
    targets: np.ndarray  # shape-(N,)
    _class_to_idx = None

    def __getitem__(self, index: int) -> Tuple[Any, Any]:
        """
        Parameters
        ----------
        index: int

        Returns
        -------
        Tuple
            (transform(image), target_transform(target)) where
            target is the index of the target class.
        """
        img, target = self.data[index], self.targets[index]

        # doing this so that it is consistent with all other datasets
        # to return a PIL Image
        img = Image.fromarray(img)

        if self.transform is not None:
            img = self.transform(img)

        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    @property
    def class_to_idx(self) -> Dict[str, int]:
        if self._class_to_idx is None:
            self._class_to_idx = {_class: i for i, _class in enumerate(self.classes)}
        return self._class_to_idx

    def __len__(self) -> int:
        return len(self.data)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# pyright: strict
import functools
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    overload,
)

import numpy as np
from torch import Tensor
from typing_extensions import Protocol, TypeAlias, TypeGuard

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)

__all__ = [
    "InstantiatesTo",
    "instantiates_to",
    "Optimizer",
    "OptimParams",
    "OptimizerType",
    "ParamGroup",
    "Partial",
    "OptimParams",
]


class Partial(Protocol[T_co]):
    def __call__(self, *args: Any, **kwargs: Any) -> T_co:  # pragma: no cover
        ...


class Optimizer(Protocol):  # pragma: no cover
    defaults: Dict[str, Any]
    state: Any
    param_groups: Any

    def __setstate__(self, state: Any) -> None:
        ...

    def state_dict(self) -> Any:
        ...

    def load_state_dict(self, state_dict: Any) -> None:
        ...

    def zero_grad(self, set_to_none: Optional[bool] = ...) -> None:
        ...

    def step(self, closure: Optional[Callable[[], Any]] = ...) -> Optional[Any]:
        ...

    def add_param_group(self, param_group: Any) -> None:
        ...


ParamGroup = Mapping[str, Any]

# Needed instead of Type[Optimizer] to deal with __init__ sig mismatch
OptimizerType: TypeAlias = Callable[..., Optimizer]


# type for Optim(params: <>)
OptimParams = Union[Sequence[Tensor], Iterable[ParamGroup]]


InstantiatesTo = Union[Type[T_co], Partial[T_co]]


def _is_protocol(cls: Any) -> bool:
    return issubclass(cls, Generic) and getattr(cls, "_is_protocol", False)


# This should *only* be used internally and with care.
# There are complications for using this with protocols.
def instantiates_to(x: Any, co_var_type: Type[T]) -> TypeGuard[InstantiatesTo[T]]:
    """Checks if `x(...)` will return type `co_var_type`.

    Accomodates structural subtyping via protocols.

    Parameters
    ----------
    x : Any

    co_var_type : Type[T]
        A type or a protocol. Note that protocols should not
        include instance-level attributes.

    Returns
    -------
    TypeGuard[InstantiatesTo[T]]
        True if `isinstance(x(...), co_var_type)` would return True."""
    assert isinstance(co_var_type, type), f"`type_` must be a type, got {co_var_type}"

    if inspect.isclass(x):
        obj = x
    elif isinstance(x, functools.partial):
        obj = x.func
        if not inspect.isclass(obj):
            return False
    else:
        return False

    return (
        isinstance(obj, co_var_type)
        if _is_protocol(co_var_type)
        else issubclass(obj, co_var_type)
    )


Scalar = Union[
    int,
    float,
    complex,
    str,
    bytes,
    np.generic,
]


class _SupportsArray(Protocol):  # pragma: no cover
    @overload
    def __array__(self, __dtype: Any = ...) -> np.ndarray:  # type: ignore
        ...

    @overload
    def __array__(self, dtype: Any = ...) -> np.ndarray:  # type: ignore
        ...


# mypy doesn't support recursive types
ArrayLike = Union[
    Scalar,
    Sequence[Scalar],
    Sequence[Sequence[Any]],
    _SupportsArray,
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import random
from collections import defaultdict
from typing import (
    Any,
    Callable,
    Collection,
    DefaultDict,
    Dict,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import numpy as np
import torch as tr
from torch.nn.functional import softmax
from torchmetrics import Metric

from rai_toolbox import evaluating
from rai_toolbox._utils import get_device

from ._fourier_basis import generate_fourier_bases

T = TypeVar("T", np.ndarray, tr.Tensor)


def perturb_batch(
    *,
    batch: T,
    basis: np.ndarray,
    basis_norm: float,
    rand_flip_per_channel: bool,
) -> T:
    """
    Given a single Fourier basis array, perturbs a batch of images by applying
        color_channel += rand_sign * norm * normed_basis

    to each color channel of each image.

    Parameters
    ----------
    batch: np.ndarray, shape-(N, C, H, W)
        N: batch size, C: number of color channels.

    basis: np.ndarray, shape-(H, W)
        Assumed to already be normalized

    basis_norm: float

    rand_flip_per_channel: bool
        If True, the std-lib `random` module is used to draw a random sign associated
        with each of the to-be-perturbed color channels.

    Returns
    -------
    perturbed_batch: np.ndarray, shape-(N, C, H, W)
    """
    basis = basis * basis_norm

    if not rand_flip_per_channel:
        if isinstance(batch, tr.Tensor):
            basis = tr.tensor(basis, dtype=batch.dtype, device=batch.device)
        return batch + basis

    _channel_flips = np.array(
        [random.randrange(-1, 2, 2) for i in range(batch.shape[0] * batch.shape[1])],
        dtype="float32",
    )

    out = np.multiply.outer(_channel_flips, basis).reshape(batch.shape)

    if isinstance(batch, tr.Tensor):
        out = tr.tensor(out, dtype=batch.dtype, device=batch.device)

    out += batch
    return out


def normalize(
    imgs: np.ndarray,
    *,
    mean: Union[float, Sequence[float]],
    std: Union[float, Sequence[float]],
    inplace: bool = False,
):
    """
    Returns (imgs - mean) / std

    Parameters
    ----------
    imgs: np.ndarray, shape-(N, C, H, W)
    mean : float | shape-(C,)
    std : float | shape-(C,)
    inplace: bool, optional (default=False)

    Returns
    -------
    normalized_imgs: np.ndarray, shape-(N, C, H, W)
    """
    mean_arr = np.atleast_1d(mean)[None, :, None, None].astype(imgs.dtype)
    std_arr = np.atleast_1d(std)[None, :, None, None].astype(imgs.dtype)

    assert imgs.ndim == 4
    if not inplace:
        return (imgs - mean_arr) / std_arr
    else:
        imgs -= mean_arr
        imgs /= std_arr
        return imgs


class HeatMapEntry(NamedTuple):
    pos: Tuple[int, int]
    sym_pos: Tuple[int, int]
    result: Any


def create_heatmaps(
    dataloader: Collection[Tuple[tr.Tensor, tr.Tensor]],
    image_height_width: Tuple[int, int],
    *,
    model: tr.nn.Module,
    metrics: Dict[str, Type[Metric]],
    basis_norm: float,
    rand_flip_per_channel: bool,
    post_pert_batch_transform: Optional[Callable[[tr.Tensor], tr.Tensor]] = None,
    device: Optional[Union[tr.device, str, int]] = None,
    row_col_coords: Optional[Iterable[Tuple[int, int]]] = None,
    factor_2pi_phase_shift: float = 0,
) -> Dict[str, List[HeatMapEntry]]:
    from rai_toolbox._utils.tqdm import tqdm
    
    _outer_total = (
        None if row_col_coords is not None else int(np.prod(image_height_width)) // 2
    )

    with evaluating(model), tr.no_grad():
        if device is not None:
            device = tr.device(device)
            model.to(device=device)
        else:
            device = get_device(model)

        if post_pert_batch_transform is None:

            def post_pert_batch_transform(x):
                return x

        results: DefaultDict[str, Dict[Tuple[int, int], HeatMapEntry]] = defaultdict(
            dict
        )

        for batch, targets in tqdm(dataloader, desc="batch"):
            batch = batch.pin_memory()
            targets = targets.to(device=device)

            for basis in tqdm(
                generate_fourier_bases(
                    *image_height_width,
                    dtype="float32",
                    row_col_coords=row_col_coords,
                    factor_2pi_phase_shift=factor_2pi_phase_shift,
                ),
                total=_outer_total,
                desc="fourier-grid",
                leave=False,
            ):
                p_batch = batch.to(device=device)

                p_batch = perturb_batch(
                    batch=p_batch,
                    basis=basis.basis,
                    basis_norm=basis_norm,
                    rand_flip_per_channel=rand_flip_per_channel,
                )

                p_batch = post_pert_batch_transform(p_batch)

                probs = softmax(model(p_batch), dim=1).cpu()
                targets = targets.cpu()

                for name, M in metrics.items():
                    if basis.position not in results[name]:
                        results[name][basis.position] = HeatMapEntry(
                            basis.position,
                            sym_pos=basis.sym_position,
                            result=M(),
                        )
                    results[name][basis.position].result.update(
                        preds=probs, target=targets
                    )

        out: Dict[str, List[HeatMapEntry]] = {}
        for metric_name, r in results.items():
            out[metric_name] = [
                HeatMapEntry(p, ps, metric.compute())
                for _, (p, ps, metric) in r.items()
            ]
        return out



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from abc import ABC, abstractmethod, abstractstaticmethod
from collections import UserList, defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch as tr
from hydra.core.utils import JobReturn
from hydra_zen import load_from_yaml, make_config
from typing_extensions import Self, TypeAlias, TypeGuard

from rai_toolbox._utils import value_check

from .hydra import launch, zen

LoadedValue: TypeAlias = Union[str, int, float]

__all__ = [
    "BaseWorkflow",
    "RobustnessCurve",
    "MultiRunMetricsWorkflow",
    "multirun",
    "hydra_list",
]


class multirun(UserList):
    """Signals that a sequence is to be iterated over in a multirun"""

    pass


class hydra_list(UserList):
    """Signals that a sequence is provided as a single configured value (i.e. it is not
    to be iterated over during a multirun)"""

    pass


class BaseWorkflow(ABC):
    """Provides an interface for creating a reusable workflow: encapsulated
    "boilerplate" for running, aggregating, and analyzing one or more Hydra jobs.

    Attributes
    ----------
    cfgs : List[Any]
        List of configurations for each Hydra job.

    metrics : Dict[str, List[Any]]
        Dictionary of metrics for across all jobs.

    workflow_overrides : Dict[str, Any]
        Workflow parameters defined as additional arguments to `run`.

    jobs : List[Any]
        List of jobs returned for each experiment within the workflow.

    working_dir: pathlib.Path
        The working directory of the experiment defined by Hydra's sweep directory
        (`hydra.sweep.dir`).
    """

    cfgs: List[Any]
    metrics: Dict[str, List[Any]]
    workflow_overrides: Dict[str, Any]
    jobs: List[Any]
    working_dir: Path

    def __init__(self, eval_task_cfg=None) -> None:
        """Workflows and experiments using Hydra.

        Parameters
        ----------
        eval_task_cfg: Mapping | None (default: None)
            The workflow configuration object.

        """
        # we can do validation checks here
        self.eval_task_cfg = (
            eval_task_cfg if eval_task_cfg is not None else make_config()
        )

        # initialize attributes
        self.cfgs = []
        self.metrics = {}
        self.workflow_overrides = {}
        self._multirun_task_overrides = {}
        self.jobs = []
        self.working_dir = Path.cwd()  # TODO: I don't think we should assume this

    @property
    def multirun_task_overrides(
        self,
    ) -> Dict[str, Union[LoadedValue, Sequence[LoadedValue]]]:
        # e.g. {'epsilon': [1.0, 2.0, 3.0], "foo": "apple"}
        if not self._multirun_task_overrides:
            for entry in load_from_yaml(
                self.working_dir / "multirun.yaml"
            ).hydra.overrides.task:
                entry: str  # e.g. +epsilon=1.0,3.0,2.0
                param_name, val = entry.split("=", 1)
                if "," in val:
                    val = [_num_from_string(v) for v in val.split(",")]
                else:
                    val = _num_from_string(val)

                param_name = param_name.split("+")[-1]
                self._multirun_task_overrides[param_name] = val
        return self._multirun_task_overrides

    @abstractstaticmethod
    def evaluation_task(*args: Any, **kwargs: Any) -> Any:
        """User-defined evaluation task to run the workflow.

        Arguments should be instantiated configuration variables.  For example,
        if the the workflow configuration is structured as::

            ├── eval_task_cfg
            │    ├── trainer
            |    ├── module
            |    ├── another_config

        The inputs to `evaluation_task` can be any of the three configurations:
        `trainer`, `module`, or `another_config` such as::

            def evaluation_task(trainer: Trainer, module: LightningModule) -> None:
                trainer.fit(module)
        """
        raise NotImplementedError()

    def validate(self):
        """Valide the configuration will execute with the user defined evaluation task"""
        zen(self.evaluation_task).validate(self.eval_task_cfg)

    def run(
        self,
        *,
        working_dir: Optional[str] = None,
        sweeper: Optional[str] = None,
        launcher: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        **workflow_overrides: Union[str, int, float, bool, dict, multirun, hydra_list],
    ):
        """Run the experiment.

        Individual workflows can expclitly define `workflow_overrides` to improve
        readability and undstanding of what parameters are expected for a particular
        workflow.

        Parameters
        ----------
        working_dir: str (default: None, the Hydra default will be used)
            The directory to run the experiment in.  This value is used for
            setting `hydra.sweep.dir`.

        sweeper: str | None (default: None)
            The configuration name of the Hydra Sweeper to use (i.e., the override for
            `hydra/sweeper=sweeper`)

        launcher: str | None (default: None)
            The configuration name of the Hydra Launcher to use (i.e., the override for
            `hydra/launcher=launcher`)

        overrides: List[str] | None (default: None)
            Parameter overrides not considered part of the workflow parameter set.
            This is helpful for filtering out parameters stored in
            `self.workflow_overrides`.

        **workflow_overrides: str | int | float | bool | multirun | hydra_list | dict
            These parameters represent the values for configurations to use for the
            experiment.

            Passing `param=multirun([1, 2, 3])` will perform a multirun over those
            three param values, whereas passing `param=hydra_list([1, 2, 3])` will
            pass the entire list as a single input.

            These values will be appended to the `overrides` for the Hydra job.
        """
        self._workflow_overrides = workflow_overrides

        if overrides is None:
            overrides = []

        if working_dir is not None:
            overrides.append(f"hydra.sweep.dir={working_dir}")

        if sweeper is not None:
            overrides.append(f"hydra/sweeper={sweeper}")

        if launcher is not None:
            overrides.append(f"hydra/launcher={launcher}")

        for k, v in workflow_overrides.items():
            value_check(k, v, type_=(int, float, bool, str, dict, multirun, hydra_list))
            if isinstance(v, multirun):
                v = ",".join(str(item) for item in v)

            prefix = ""
            if (
                not hasattr(self.eval_task_cfg, k)
                or getattr(self.eval_task_cfg, k) is None
            ):
                prefix = "+"

            overrides.append(f"{prefix}{k}={v}")

        # Run a Multirun over epsilons
        (jobs,) = launch(
            self.eval_task_cfg,
            zen(self.evaluation_task),
            overrides=overrides,
            multirun=True,
        )

        self.jobs = jobs
        self.jobs_post_process()

    @abstractmethod
    def jobs_post_process(self):
        """Method to extract attributes and metrics relevant to the workflow."""

    def plot(self, **kwargs) -> None:
        """Plot workflow metrics."""
        raise NotImplementedError()

    def to_dataframe(self):
        """Convert workflow data to Pandas DataFrame."""
        raise NotImplementedError()

    def to_xarray(self):
        """Convert workflow data to xArray Dataset or DataArray."""
        raise NotImplementedError()


def _num_from_string(str_input: str) -> LoadedValue:
    try:
        val = float(str_input)
        if val.is_integer() and "." not in str_input:
            # v is e.g., 1 or -2. Not 1.0 or -2.0
            val = int(str_input)
        return val
    except ValueError:
        return str_input


def _non_str_sequence(x: Any) -> TypeGuard[Sequence[Any]]:
    return isinstance(x, Sequence) and not isinstance(x, str)


class MultiRunMetricsWorkflow(BaseWorkflow):
    """Abstract class for workflows that record metrics using Hydra multirun.

    This workflow creates subdirectories of multirun experiments using Hydra.  These directories
    contain the Hydra YAML configuration and any saved metrics file (defined by the evaulation task)::

        ├── working_dir
        │    ├── <experiment directory name: 0>
        │    |    ├── <hydra output subdirectory: (default: .hydra)>
        |    |    |    ├── config.yaml
        |    |    |    ├── hydra.yaml
        |    |    |    ├── overrides.yaml
        │    |    ├── <metrics_filename>
        │    ├── <experiment directory name: 1>
        |    |    ...

    The evaluation task is expected to return a dictionary that maps
    `metric-name (str) -> value (number | Sequence[number])`

    Examples
    --------
    Let's create a simple workflow where we perform a multirun over a parameter,
    `epsilon`, and evaluate a task function that computes an accuracy and loss based on
    that `epsilon` value and a specified `scale`.

    >>> from rai_toolbox.mushin.workflows import MultiRunMetricsWorkflow
    >>> from rai_toolbox.mushin import multirun
    >>> import torch as tr
    ... class LocalRobustness(MultiRunMetricsWorkflow):
    ...     @staticmethod
    ...     def evaluation_task(epsilon: float, scale: float) -> dict:
    ...         epsilon *= scale
    ...         val = 100 - epsilon**2
    ...         result = dict(accuracies=val+2, loss=epsilon**2)
    ...         tr.save(result, "test_metrics.pt")
    ...         return result

    We'll run this workflow for six total configurations of three `epsilon` values and two `scale` values. This will launch a Hydra multirun job and aggregate the results.

    >>> wf = LocalRobustness()
    >>> wf.run(epsilon=multirun([1.0, 2.0, 3.0]), scale=multirun([0.1, 1.0]))
    [2022-05-02 11:57:59,219][HYDRA] Launching 6 jobs locally
    [2022-05-02 11:57:59,220][HYDRA] 	#0 : +epsilon=1.0 +scale=0.1
    [2022-05-02 11:57:59,312][HYDRA] 	#1 : +epsilon=1.0 +scale=1.0
    [2022-05-02 11:57:59,405][HYDRA] 	#2 : +epsilon=2.0 +scale=0.1
    [2022-05-02 11:57:59,498][HYDRA] 	#3 : +epsilon=2.0 +scale=1.0
    [2022-05-02 11:57:59,590][HYDRA] 	#4 : +epsilon=3.0 +scale=0.1
    [2022-05-02 11:57:59,683][HYDRA] 	#5 : +epsilon=3.0 +scale=1.0

    Now that this workflow has run, we can view the results as an xarray-dataset whose
    coordinates reflect the multirun parameters that were varied, and whose data-variables are our recorded metrics: "accuracies" and "loss".

    >>> ds = wf.to_xarray()
    >>> ds
    <xarray.Dataset>
    Dimensions:     (epsilon: 3, scale: 2)
    Coordinates:
    * epsilon     (epsilon) float64 1.0 2.0 3.0
    * scale       (scale) float64 0.1 1.0
    Data variables:
        accuracies  (epsilon, scale) float64 102.0 101.0 102.0 98.0 101.9 93.0
        loss        (epsilon, scale) float64 0.01 1.0 0.04 4.0 0.09 9.0

    We can also load this workflow by providing the working directory where it was run.

    >>> loaded = LocalRobustness().load_from_dir(wf.working_dir)
    >>> loaded.to_xarray()
    <xarray.Dataset>
    Dimensions:     (epsilon: 3, scale: 2)
    Coordinates:
    * epsilon     (epsilon) float64 1.0 2.0 3.0
    * scale       (scale) float64 0.1 1.0
    Data variables:
        accuracies  (epsilon, scale) float64 102.0 101.0 102.0 98.0 101.9 93.0
        loss        (epsilon, scale) float64 0.01 1.0 0.04 4.0 0.09 9.0
    """

    @abstractstaticmethod
    def evaluation_task(*args: Any, **kwargs: Any) -> Dict[str, Any]:
        raise NotImplementedError()

    def jobs_post_process(self):
        assert len(self.jobs) > 0
        # TODO: Make protocol type for JobReturn
        assert isinstance(self.jobs[0], JobReturn)

        # set working directory of this workflow
        first_job_working_dir = self.jobs[0].working_dir
        assert first_job_working_dir is not None
        self.working_dir = Path(first_job_working_dir).parent

        # extract configs, overrides, and metrics
        self.cfgs = [j.cfg for j in self.jobs]
        job_overrides = [j.hydra_cfg.hydra.overrides.task for j in self.jobs]
        job_metrics = [j.return_value for j in self.jobs]
        workflow_params = list(self._workflow_overrides.keys())

        self.metrics, self.workflow_overrides = _load_metrics(
            job_overrides, job_metrics, workflow_params
        )

    def load_from_dir(
        self: Self,
        working_dir: Union[Path, str],
        config_dir: str = ".hydra",
        metrics_filename: str = "test_metrics.pt",
        workflow_params: Optional[Sequence[str]] = None,
    ) -> Self:
        """Loading workflow job data from a given working directory. The workflow
        is loaded in-place and "self" is returned by this method.

        Parameters
        ----------
        working_dir: str | Path
            The base working directory of the experiment. It is expected
            that subdirectories within this working directory will contain
            individual Hydra jobs data (yaml configurations) and saved metrics files.

        config_dir: str (default: ".hydra")
            The directory in an experiment that stores Hydra configurations (`hydra.output_subdir`)

        metrics_filename: str (default: "test_metrics.pt")
            The filename used to save metrics for each individual Hydra job. This can
            be a search pattern as well since this is appended to

                `Path(working_dir).glob("**/*/<metrics_filename")`

        workflow_params: Sequence[str] | None (default: None)
            A string of parameters to use for `workflow_params`.  If `None` it will
            default to all parameters saved in Hydra's `overrides.yaml` file.

        Returns
        -------
        loaded_workflow : Self
        """
        self.working_dir = Path(working_dir).resolve()

        multirun_cfg = self.working_dir / "multirun.yaml"
        assert (
            multirun_cfg.exists()
        ), "Working directory does not contain `multirun.yaml` file.  Be sure to use the value of the Hydra sweep directory for the workflow"

        # Load saved YAML configurations for each job (in hydra.job.output_subdir)
        job_cfgs = [
            load_from_yaml(f)
            for f in sorted(self.working_dir.glob(f"**/*/{config_dir}/config.yaml"))
        ]

        # Load saved YAML overrides for each job (in hydra.job.output_subdir)
        job_overrides = [
            list(load_from_yaml(f))
            for f in sorted(self.working_dir.glob(f"**/*/{config_dir}/overrides.yaml"))
        ]

        # Load metrics for each job
        job_metrics = [
            tr.load(f)
            for f in sorted(self.working_dir.glob(f"**/*/{metrics_filename}"))
        ]

        self.cfgs = job_cfgs
        self.metrics, self.workflow_overrides = _load_metrics(
            job_overrides, job_metrics, workflow_params
        )
        return self

    def to_dataframe(self):
        """Convert workflow data to Pandas DataFrame."""
        import pandas as pd

        d = {}
        d.update(self.metrics)
        d.update(self.workflow_overrides)
        return pd.DataFrame(d)

    def to_xarray(self, non_multirun_params_as_singleton_dims: bool = False):
        """Convert workflow data to xarray Dataset.

        Parameters
        ----------
        non_multirun_params_as_singleton_dims : bool, optional (default=False)
            If `True` then non-multirun entries from `workflow_overrides` will be
            included as length-1 dimensions in the xarray. Useful for merging/
            concatenation with other Datasets

        Returns
        -------
        results : xarray.Dataset
            A dataset whose dimensions and coordinate-values are determined by the
            quantities over which the multi-run was performed. The data variables correspond to the named results returned by the jobs."""
        import xarray as xr

        orig_coords = {
            k: (v if _non_str_sequence(v) else [v])
            for k, v in self.multirun_task_overrides.items()
            if non_multirun_params_as_singleton_dims or _non_str_sequence(v)
        }

        # non-multirun overrides
        attrs = {
            k: v
            for k, v in self.multirun_task_overrides.items()
            if not _non_str_sequence(v)
        }

        # we will add additional coordinates as-needed for multi-dim metrics
        coords: Dict[str, Any] = orig_coords.copy()
        shape = tuple(len(v) for v in coords.values())

        data = {}
        for k, v in self.metrics.items():
            datum = np.asarray(v).reshape(shape + np.asarray(v[0]).shape)
            n = 0
            for n in range(datum.ndim - len(coords)):
                # Create additional arbitrary coordinates as-needed for non-scalar
                # metrics
                coords[f"{k}_dim{n}"] = np.arange(datum.shape[len(orig_coords) + n])

            data[k] = (
                list(orig_coords)
                + [f"{k}_dim{i}" for i in range(datum.ndim - len(shape))],
                datum,
            )
        return xr.Dataset(coords=coords, data_vars=data, attrs=attrs)


class RobustnessCurve(MultiRunMetricsWorkflow):
    """Abstract class for workflows that measure performance for different perturbation values.

    This workflow requires and uses parameter `epsilon` as the configuration option for varying the perturbation.

    See Also
    --------
    MultiRunMetricsWorkflow
    """

    def run(
        self,
        *,
        epsilon: Union[str, Sequence[float]],
        working_dir: Optional[str] = None,
        sweeper: Optional[str] = None,
        launcher: Optional[str] = None,
        overrides: Optional[List[str]] = None,
        **workflow_overrides: Union[str, int, float, bool, multirun, hydra_list],
    ):
        """Run the experiment for varying value `epsilon`.

        Parameters
        ----------
        epsilon: str | Sequence[float]
            The configuration parameter for the perturbation.  Unlike Hydra overrides,
            this parameter can be a list of floats that will be converted into a
            multirun sequence override for Hydra.

        working_dir: str (default: None, the Hydra default will be used)
            The directory to run the experiment in.  This value is used for
            setting `hydra.sweep.dir`.

        sweeper: str | None (default: None)
            The configuration name of the Hydra Sweeper to use (i.e., the override for
            `hydra/sweeper=sweeper`)

        launcher: str | None (default: None)
            The configuration name of the Hydra Launcher to use (i.e., the override for
            `hydra/launcher=launcher`)

        overrides: List[str] | None (default: None)
            Parameter overrides not considered part of the workflow parameter set.
            This is helpful for filtering out parameters stored in
            `self.workflow_overrides`.

        **workflow_overrides: dict | str | int | float | bool | multirun | hydra_list
            These parameters represent the values for configurations to use for the
            experiment.

            These values will be appeneded to the `overrides` for the Hydra job.
        """

        if not isinstance(epsilon, str):
            epsilon = multirun(epsilon)

        return super().run(
            working_dir=working_dir,
            sweeper=sweeper,
            launcher=launcher,
            overrides=overrides,
            **workflow_overrides,
            # for multiple multi-run params, epsilon should fastest-varying param;
            # i.e. epsilon should be the trailing dim in the multi-dim array of results
            epsilon=epsilon,
        )

    def to_dataframe(self):
        """Convert workflow data to Pandas DataFrame."""
        return super().to_dataframe().sort_values("epsilon")

    def to_xarray(self, non_multirun_params_as_singleton_dims: bool = False):
        """Convert workflow data to xarray Dataset.

        Parameters
        ----------
        non_multirun_params_as_singleton_dims : bool, optional (default=False)
            If `True` then non-multirun entries from `workflow_overrides` will be
            included as length-1 dimensions in the xarray. Useful for merging/
            concatenation with other Datasets

        Returns
        -------
        results : xarray.Dataset
            A dataset whose dimensions and coordinate-values are determined by the
            quantities over which the multi-run was performed. The data variables correspond to the named results returned by the jobs."""
        return (
            super()
            .to_xarray(
                non_multirun_params_as_singleton_dims=non_multirun_params_as_singleton_dims
            )
            .sortby("epsilon")
        )

    def plot(
        self,
        metric: str,
        ax=None,
        group: Optional[str] = None,
        save_filename: Optional[str] = None,
        non_multirun_params_as_singleton_dims: bool = False,
        **kwargs,
    ) -> None:
        """Plot metrics versus `epsilon`.

        Using the `xarray.Dataset` from `to_xarray`, plot the metrics
        against the workflow perturbation parameters.

        Parameters
        ----------
        metric: str
            The metric saved

        ax: Axes | None (default: None)
            If not `None`, the matplotlib.Axes to use for plotting.

        group: str | None (default: None)
            Needed if other parameters besides `epsilon` were varied.

        save_filename: str | None (default: None)
            If not `None` save figure to the filename provided.

        non_multirun_params_as_singleton_dims : bool, optional (default=False)
            If `True` then non-multirun entries from `workflow_overrides` will be
            included as length-1 dimensions in the xarray. Useful for merging/
            concatenation with other Datasets

        **kwargs: Any
            Additional arguments passed to `xarray.plot`.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()

        xdata = self.to_xarray(
            non_multirun_params_as_singleton_dims=non_multirun_params_as_singleton_dims
        )
        if group is None:
            plots = xdata[metric].plot.line(x="epsilon", ax=ax, **kwargs)

        else:
            # TODO: xarray.groupby doesn't support multidimensional grouping
            dg = xdata.groupby(group)
            plots = [
                grp[metric].plot(x="epsilon", label=name, ax=ax, **kwargs)
                for name, grp in dg
            ]

        if save_filename is not None:
            plt.savefig(save_filename)

        return plots


def _load_metrics(
    job_overrides: List[List[str]],
    job_metrics: List[Dict[str, Any]],
    workflow_params: Optional[Sequence[str]] = None,
) -> Tuple[Dict[str, List[Any]], Dict[str, List[Any]]]:
    workflow_overrides = defaultdict(list)
    metrics = defaultdict(list)
    for task_overrides, task_metrics in zip(job_overrides, job_metrics):
        for override in task_overrides:
            k, v = override.split("=")
            # remove any hydra override prefix
            param = k.split("+")[-1]
            if workflow_params is None or param in workflow_params:
                v = _num_from_string(v)
                workflow_overrides[param].append(v)

        for k, v in task_metrics.items():
            # get item if it's a single element array
            if isinstance(v, list) and len(v) == 1:
                v = v[0]

            metrics[k].append(v)

    return metrics, workflow_overrides



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from setuptools import find_packages, setup

import versioneer

DISTNAME = "rai_toolbox"
LICENSE = "MIT"
AUTHOR = "Ryan Soklaski, Justin Goodwin, Olivia Brown, Michael Yee"
AUTHOR_EMAIL = "ryan.soklaski@ll.mit.edu"
URL = "https://github.com/mit-ll-responsible-ai/responsible-ai-toolbox"
CLASSIFIERS = [
    "Development Status :: 4 - Beta",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Intended Audience :: Science/Research",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Topic :: Scientific/Engineering",
]
KEYWORDS = "machine learning robustness pytorch responsible AI"
INSTALL_REQUIRES = [
    "numpy > 1.17.2, != 1.19.3",
    "torchvision >= 0.10.0",
    "torch >= 1.9.0",
    "torchmetrics >= 0.6.0",
    "typing-extensions >= 4.1.1",
]
TESTS_REQUIRE = [
    "pytest >= 3.8",
    "hypothesis >= 6.41.0",
    "mygrad >= 2.0.0",
    "omegaconf >= 2.1.1",
]

DESCRIPTION = "PyTorch-centric library for evaluating and enhancing the robustness of AI technologies"
LONG_DESCRIPTION = """
The rAI-toolbox is designed to enable methods for evaluating and enhancing both the
robustness and the explainability of AI models in a way that is scalable and that
composes naturally with other popular ML frameworks.

A key design principle of the rAI-toolbox is that it adheres strictly to the APIs
specified by the PyTorch machine learning framework. For example, the rAI-toolbox frames
adversarial training workflows solely in terms of the `torch.nn.Optimizer` and
`torch.nn.Module` APIs. This makes it trivial to leverage other libraries and
frameworks from the PyTorch ecosystem to bolster your responsible AI R&D. For
instance, one can naturally leverage the rAI-toolbox together with
PyTorch Lightning to perform distributed adversarial training.
"""


setup(
    name=DISTNAME,
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    license=LICENSE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    install_requires=INSTALL_REQUIRES,
    tests_require=TESTS_REQUIRE,
    url=URL,
    download_url=f"{URL}/tarball/v" + versioneer.get_version(),
    python_requires=">=3.7",
    packages=find_packages(where="src", exclude=["tests", "tests.*"]),
    package_dir={"": "src"},
    package_data={"rai_toolbox": ["py.typed"]},
    extras_require={
        "tests": TESTS_REQUIRE,
        "mushin": [
            "pytorch-lightning >= 1.5.0",
            "hydra-zen >= 0.6.0",
            "xarray >= 0.19.0",
            "matplotlib >= 3.3",
        ],
    },
)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import NamedTuple

import pytorch_lightning
from typing_extensions import Final


class Version(NamedTuple):
    major: int
    minor: int
    patch: int


def _get_version(ver_str: str) -> Version:
    # Not for general use. Tested only for Hydra and OmegaConf
    # version string styles

    splits = ver_str.split(".")[:3]
    if not len(splits) == 3:  # pragma: no cover
        raise ValueError(f"Version string {ver_str} couldn't be parsed")

    major, minor = (int(v) for v in splits[:2])
    patch_str, *_ = splits[-1].split("rc")

    return Version(major=major, minor=minor, patch=int(patch_str))


PL_VERSION: Final = _get_version(pytorch_lightning.__version__)



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from .transforms import AugMix, Fork, augment_and_mix

__all__ = ["AugMix", "Fork", "augment_and_mix"]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from .frank_wolfe import (
    FrankWolfe,
    L1FrankWolfe,
    L1qFrankWolfe,
    L2FrankWolfe,
    LinfFrankWolfe,
)
from .lp_space import (
    L1NormedGradientOptim,
    L1qNormedGradientOptim,
    L2NormedGradientOptim,
    L2ProjectedOptim,
    LinfProjectedOptim,
    SignedGradientOptim,
)
from .misc import (
    ClampedGradientOptimizer,
    ClampedParameterOptimizer,
    TopQGradientOptimizer,
)
from .optimizer import ChainedParamTransformingOptimizer, ParamTransformingOptimizer

__all__ = [
    "L1FrankWolfe",
    "L1qFrankWolfe",
    "L2FrankWolfe",
    "LinfFrankWolfe",
    "FrankWolfe",
    "L1NormedGradientOptim",
    "L1qNormedGradientOptim",
    "L2NormedGradientOptim",
    "L2ProjectedOptim",
    "SignedGradientOptim",
    "LinfProjectedOptim",
    "ParamTransformingOptimizer",
    "TopQGradientOptimizer",
    "ClampedGradientOptimizer",
    "ClampedParameterOptimizer",
    "ChainedParamTransformingOptimizer",
]



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import functools
from typing import Any, Callable, List, Optional, Tuple, Union

import torch as tr
from torch import Tensor
from torch.nn import CrossEntropyLoss, Module
from torch.optim import Optimizer as _TorchOptim

from rai_toolbox import negate
from rai_toolbox._typing import (
    ArrayLike,
    InstantiatesTo,
    Optimizer,
    OptimizerType,
    Partial,
    instantiates_to,
)
from rai_toolbox._utils import value_check
from rai_toolbox._utils.stateful import evaluating, frozen
from rai_toolbox.perturbations import AdditivePerturbation, PerturbationModel


def gradient_ascent(
    *,
    model: Callable[[Tensor], Tensor],
    data: ArrayLike,
    target: ArrayLike,
    optimizer: Union[Optimizer, OptimizerType, Partial[Optimizer]],
    steps: int,
    perturbation_model: Union[
        PerturbationModel, InstantiatesTo[PerturbationModel]
    ] = AdditivePerturbation,
    targeted: bool = False,
    use_best: bool = False,
    criterion: Optional[Callable[[Tensor, Tensor], Tensor]] = None,
    reduction_fn: Callable[[Tensor], Tensor] = tr.sum,
    **optim_kwargs: Any,
) -> Tuple[Tensor, Tensor]:
    """Solve for a set of perturbations for a given set of data and a model,
    and then apply those perturbations to the data.

    This performs, for `steps` iterations, the following optimization::

       optim = optimizer(perturbation_model.parameters)
       pert_data = perturbation_model(data)
       loss = criterion(model(pert_data), target)
       loss = (1 if targeted else -1) * loss  # default: targeted=False
       optim.step()

    Note that, by default, this perturbs the data away from `target` (i.e., this performs
    gradient *ascent*), given a standard loss function that seeks to minimize the
    diffence between the model's output and the target. See `targeted` to toggle this
    behavior.

    Parameters
    ----------
    model : Callable[[Tensor], Tensor]
        Differentiable function that processes the (perturbed) data prior to computing
        the loss.

        If `model` is a `torch.nn.Module`, then its weights will be frozen and it will
        be set to eval mode during the perturbation-solve phase.

    data : Tensor, shape-(N, ...)
        The input data to perturb.

    target : Tensor, shape-(N, ...)
        If `targeted==False` (default), then this is the target to perturb away from.
        If `targeted==True`, then this is the target to perturb toward.

    optimizer : Optimizer | Type[Optimizer] | Partial[Optimizer]
        The optimizer to use for updating the perturbation model.

        If `optimizer` is uninstantiated, it will be instantiated as `optimizer(perturbation_model.parameters(), **optim_kwargs)`

    steps : int
        Number of projected gradient steps.

    perturbation_model : PerturbationModel | Type[PerturbationModel], optional (default=AdditivePerturbation)
        A `torch.nn.Module` whose parameters are updated by the solver. Its forward-pass applies the perturbation to the data. Default is
        `AdditivePerturbation`, which simply adds the perturbation to the data.

        The perturbation model should not modify the data in-place.

        If `perturbation_model` is a type, then it will be instantiated as
        `perturbation_model(data)`.

    criterion : Optional[Callable[[Tensor, Tensor], Tensor]]
        The criterion to use for calculating the loss **per-datum**.  I.e., for a
        shape-(N, ...) batch of data, `criterion` should return a shape-(N,) tensor of
        loss values – one for each datum in the batch.

        If `None`, then `CrossEntropyLoss(reduction=None)` is used.

    targeted : bool (default: False)
        If `True`, then perturb towards the defined `target`, otherwise move away from
        `target`.

        Note: Default (`targeted=False`) implements gradient *ascent*.
        To perform gradient *descent*, set `targeted=True`.

    use_best : bool (default: True)
        Whether to only report the best perturbation over all steps.
        Note: Requires criterion to output a loss per sample, e.g., set
        `reduction="none"`.

    reduction_fn : Callable[[Tensor], Tensor], optional (default=torch.sum)
        Used to reduce the shape-(N,) per-datum loss to a scalar. This should be
        set to `torch.mean` when solving for a "universal" perturbation.

    **optim_kwargs : Any
       Keyword arguments passed to `optimizer` when it is instatiated.

    Returns
    -------
    xadv, losses : tuple[Tensor, Tensor], shape-(N, ...), shape-(N, ...)
        The perturbed data, if `use_best==True` then this is the best perturbation
        based on the loss across all steps.

        The loss for each perturbed data point, if `use_best==True` then this is the
        best loss across all steps.

    Notes
    -----
    `model` is automatically set to eval-mode and its parameters are set to `requires_grad=False` within the context of this function.

    Examples
    --------
    Let's perturb two data points, `x1=-1.0` and `x2=2.0`, to maximize
    `L(δ; x) = |x + δ|` w.r.t `δ`. We will use five standard gradient steps, using a
    learning  rate of 0.1. The default perturbation model is simply additive:
    `x -> x + δ`.

    This solver is refining `δ1` and `δ2`, whose initial values are 0 by default, to
    maximize `L(x) = |x|` for `x1` and `x2`, respectively. Thus we should find that our
    solved perturbations modify our data as: `x-1.0 -> -1.5` and `2.0 -> 2.5`,
    respectively.

    >>> from rai_toolbox.perturbations import gradient_ascent
    >>> from torch.optim import SGD

    >>> identity_model = lambda data: data
    >>> abs_diff = lambda model_out, target: (model_out - target).abs()

    >>> perturbed_data, losses = gradient_ascent(
    ...    data=[-1.0, 2.0],
    ...    target=0.0,
    ...    model=identity_model,
    ...    criterion=abs_diff,
    ...    optimizer=SGD,
    ...    lr=0.1,
    ...    steps=5,
    ... )
    >>> perturbed_data
    tensor([-1.5000,  2.5000])

    We can instead specify `targeted=True` and perform gradient *descent*.
    Here, the perturbations we solve for should modify our data as:
    -1.0 -> -0.5 and 2.0 -> 1.5, respectively.

    >>> perturbed_data, losses = gradient_ascent(
    ...    data=[-1.0, 2.0],
    ...    target=0.0,
    ...    model=identity_model,
    ...    criterion=abs_diff,
    ...    optimizer=SGD,
    ...    lr=0.1,
    ...    steps=5,
    ...    targeted=True,
    ... )
    >>> perturbed_data
    tensor([-0.5000,  1.5000])

    **Accessing the perturbations**

    To gain direct access to the solved perturbations, we can provide our own
    perturbation model to the solver. Let's solve the same optimization problem, but provide our own instance of `AdditivePerturbation`

    >>> from rai_toolbox.perturbations import AdditivePerturbation
    >>> pert_model = AdditivePerturbation(data_or_shape=(2,))
    >>> perturbed_data, losses = gradient_ascent(
    ...    perturbation_model=pert_model,
    ...    data=[-1.0, 2.0],
    ...    target=0.0,
    ...    model=identity_model,
    ...    criterion=abs_diff,
    ...    optimizer=SGD,
    ...    lr=0.1,
    ...    steps=5,
    ... )
    >>> perturbed_data
    tensor([-1.5000,  2.5000])

    Now we can access the values that were solved for `δ1` and `δ2`.

    >>> pert_model.delta
    Parameter containing:
    tensor([-0.5000,  0.5000], requires_grad=True)
    """
    data = tr.as_tensor(data)
    target = tr.as_tensor(target)

    if not data.is_leaf:
        data = data.detach()

    if not target.is_leaf:
        target = target.detach()

    # Initialize
    best_loss = None
    best_x = None

    if criterion is None:
        criterion = CrossEntropyLoss(reduction="none")

    if not targeted:
        # maximize the objective function
        criterion = negate(criterion)

    if instantiates_to(perturbation_model, PerturbationModel):
        pmodel = perturbation_model(data)
    else:
        if not isinstance(perturbation_model, PerturbationModel):
            raise TypeError(
                f"`perturbation_model` must be satisfy the `PerturbationModel`"
                f" protocol, got: {perturbation_model}"
            )

        pmodel = perturbation_model

    if instantiates_to(optimizer, _TorchOptim):
        optim = optimizer(pmodel.parameters(), **optim_kwargs)
    else:
        if not isinstance(optimizer, _TorchOptim):
            raise TypeError(
                f"`optimizer` must be an instance of Optimizer or must instantiate to Optimizer; got: {optimizer}"
            )
        if instantiates_to(perturbation_model, PerturbationModel):
            raise TypeError(
                "An initialized optimizer can only be passed to the solver in combination with an intialized perturbation model."
            )
        if optim_kwargs:
            raise TypeError(
                "**optim_kwargs were specified, but the provided `optimizer` has already been instaniated"
            )

        optim = optimizer

    to_freeze: List[Any] = [data, target]

    if isinstance(model, Module):
        to_freeze.append(model)

    # don't pass non nn.Module to frozen/eval
    packed_model = (model,) if isinstance(model, Module) else ()

    # Projected Gradient Descent
    with frozen(*to_freeze), evaluating(*packed_model), tr.enable_grad():
        for _ in range(steps):
            # Calculate the gradient of loss
            xadv = pmodel(data)
            logits = model(xadv)
            losses = criterion(logits, target)
            loss = reduction_fn(losses)

            # Update the perturbation
            optim.zero_grad(set_to_none=True)
            loss.backward()
            optim.step()

            if use_best:
                if (
                    (losses.ndim == 0 and data.ndim > 0)
                    or losses.ndim > data.ndim
                    or losses.shape != data.shape[: losses.ndim]
                ):
                    raise ValueError(
                        f"`use_best=True` but `criterion` does not ouput a per-datum-loss. "
                        f"I.e. `criterion` returned a tensor of shape-{tuple(losses.shape)} for a "
                        f"batch of shape-{tuple(data.shape)}. Expected a tensor of "
                        f"shape-{(len(data),)} or greater."
                    )

                best_loss, best_x = _replace_best(losses, best_loss, xadv, best_x)

        # free up memory
        optim.zero_grad(set_to_none=True)

        # Final evalulation
        with tr.no_grad():
            xadv = pmodel(data)
            logits = model(xadv)
            losses = criterion(logits, target)

            if use_best:
                # we negate the loss when `targeted=True` so min-loss is always best
                losses, xadv = _replace_best(losses, best_loss, xadv, best_x, min=True)

    if not targeted:
        # The returned loss is always relative to the original criterion.
        # E.g. an adversarial perturbation with a "low" negated loss will yield
        # a high loss.
        losses = -1 * losses
    return xadv.detach(), losses.detach()


def random_restart(
    solver: Callable[..., Tuple[Tensor, Tensor]],
    repeats: int,
) -> Callable[..., Tuple[Tensor, Tensor]]:
    """Executes a solver function multiple times, saving out the best perturbations.

    Parameters
    ----------
    solver : Callable[..., Tuple[Tensor, Tensor]]
        The solver whose execution will be repeated.

    repeats : int
        The number of times to run `solver`

    Returns
    -------
    random_restart_fn : Callable[..., Tuple[Tensor, Tensor]]
        Wrapped function that will execute `solver` `repeats` times.

    Examples
    --------
    Let's perturb two data points, `x1=-1.0` and `x2=2.0`, to maximize
    `L(δ; x) = |x + δ|` w.r.t `δ`. Our perturbation will randomly initialize `δ1` and `δ2` and we will re-run the solver three times – retaining the best perturbation of `x1` and `x2` respectively

    >>> from functools import partial
    >>> import torch as tr
    >>> from torch.optim import SGD
    >>> from rai_toolbox.perturbations.init import uniform_like_l1_n_ball_
    >>> from rai_toolbox.perturbations import AdditivePerturbation, gradient_ascent, random_restart
    >>>
    >>> def verbose_abs_diff(model_out, target):
    ...     # used to print out loss at each solver step (for purpose of example)
    ...     out =  (model_out - target).abs()
    ...     print(out)
    ...     return out

    Configuring a peturbation model to randomly initialize the perturbations.

    >>> RNG = tr.Generator().manual_seed(0)
    >>> PertModel = partial(AdditivePerturbation, init_fn=uniform_like_l1_n_ball_, generator=RNG)

    Creating and running repeating solver.

    >>> gradient_ascent_with_restart = random_restart(gradient_ascent, 3)

    >>> perturbed_data, losses = gradient_ascent_with_restart(
    ...    perturbation_model=PertModel,
    ...    data=[-1.0, 2.0],
    ...    target=0.0,
    ...    model=lambda data: data,
    ...    criterion=verbose_abs_diff,
    ...    optimizer=SGD,
    ...    lr=0.1,
    ...    steps=1,
    ... )
    tensor([0.5037, 2.7682], grad_fn=<AbsBackward0>)
    tensor([0.6037, 2.8682])
    tensor([0.9115, 2.1320], grad_fn=<AbsBackward0>)
    tensor([1.0115, 2.2320])
    tensor([0.6926, 2.6341], grad_fn=<AbsBackward0>)
    tensor([0.7926, 2.7341])

    See that for `x1` the highest loss is `1.0115`, and for `x2` it is `2.8682`. This
    should be reflected `losses` and `perturbed_data` that were retained across the
    restarts.

    >>> losses
    tensor([1.0115, 2.8682])
    >>> perturbed_data
    tensor([-1.0115,  2.8682])
    """
    value_check("repeats", repeats, min_=1, incl_min=True)

    @functools.wraps(solver)
    def random_restart_fn(*args, **kwargs) -> Tuple[Tensor, Tensor]:
        targeted = kwargs.get("targeted", False)

        best_x = None
        best_loss = None

        for _ in range(repeats):
            # run the attack
            xadv, losses = solver(*args, **kwargs)

            # Save best loss for each data point
            best_loss, best_x = _replace_best(
                loss=losses,
                best_loss=best_loss,
                data=xadv,
                best_data=best_x,
                min=targeted,
            )

        assert best_x is not None
        assert best_loss is not None
        return best_x, best_loss

    return random_restart_fn


# A function that updates the best loss and best input
def _replace_best(
    loss: Tensor,
    best_loss: Optional[Tensor],
    data: Tensor,
    best_data: Optional[Tensor],
    min: bool = True,
) -> Tuple[Tensor, Tensor]:
    """Returns the data with the smallest (or largest) loss

    Parameters
    ----------
    loss : Tensor, shape-(N, ...)
        N: batch size

    best_loss : Optional[Tensor], shape-(N, ...)
        N: batch size

    data : Tensor, shape-(N, ...)
        N: batch size

    best_data : Optional[Tensor], shape-(N, ...)
        N: batch size

    min : bool (default: True)
        Whether best is minimum (True) or maximum (False)

    Returns
    -------
    best_loss, best_data : Tuple[Tensor, Tensor]
    """
    if best_loss is None:
        best_data = data
        best_loss = loss
    else:
        assert best_data is not None
        if min:
            replace = loss < best_loss
        else:
            replace = loss > best_loss

        best_data[replace] = data[replace]
        best_loss[replace] = loss[replace]

    return best_loss, best_data



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from typing import Iterable, Optional, TypeVar

_T = TypeVar("_T", bound=Iterable)

__all__ = ["tqdm"]

def _dummy_tqdm(
    iterable: _T,
    desc: Optional[str] = None,
    total: Optional[int] = None,
    leave=True,
    file=None,
    ncols=None,
    mininterval=0.1,
    maxinterval=10.0,
    miniters=None,
    ascii=None,
    disable=False,
    unit="it",
    unit_scale=False,
    dynamic_ncols=False,
    smoothing=0.3,
    bar_format=None,
    initial=0,
    position=None,
    postfix=None,
    unit_divisor=1000,
) -> _T:
    return iterable
        
try:
    from tqdm.auto import tqdm
except ImportError:  # pragma: no cover
    tqdm = _dummy_tqdm




# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import torch
from hydra_zen import load_from_yaml
from omegaconf import DictConfig, ListConfig
from torch import nn

log = logging.getLogger(__name__)


def load_from_checkpoint(
    model: nn.Module,
    *,
    ckpt: Optional[Union[str, Path]] = None,
    weights_key: Optional[str] = None,
    weights_key_strip: Optional[str] = None,
    model_attr: Optional[str] = None,
) -> nn.Module:
    """Load model weights.

    Parameters
    ----------
    model : Module
        The PyTorch Module

    ckpt : Optional[Union[str, Path]]
        The path to the file containing the model weights. If no path is provided
        the model will not be updated.

    weights_key : Optional[str] (default: "state_dict")
        (load_module=False) The key from the checkpoint file containing the model
        weights.

    weights_key_strip : Optional[str] (default: "model")
        (load_module=False) The prefix to remove from each weight's key prior
        to loading the module.

    model_attr : Optional[str] (default: "model")
        (load_module=False) The attribute of the module containing the `torch.nn.Module`

    Returns
    -------
    module : LightningModule
    """
    if ckpt is None:
        return model

    ckpt = Path(str(ckpt))
    if not ckpt.exists():
        ckpt = Path.home() / ".torch" / "models" / ckpt
    log.info(f"Loading model checkpoint from {ckpt}")

    ckpt_data: Dict[str, Any] = torch.load(ckpt, map_location="cpu")

    if weights_key is not None:
        assert weights_key in ckpt_data
        ckpt_data = ckpt_data[weights_key]

    if weights_key_strip:
        if not weights_key_strip.endswith("."):
            weights_key_strip = weights_key_strip + "."

        ckpt_data = {
            k[len(weights_key_strip) :]: v
            for k, v in ckpt_data.items()
            if k.startswith(weights_key_strip)
        }

    if model_attr is None:
        # The weights can be loaded in directly
        model.load_state_dict(ckpt_data)  # type: ignore

    else:
        assert hasattr(model, model_attr)
        getattr(model, model_attr).load_state_dict(ckpt_data)

    return model


@dataclass
class Experiment:
    working_dir: str
    cfg: Optional[Union[Dict, ListConfig, DictConfig]]
    ckpts: List[str]
    metrics: Dict


def load_experiment(
    exp_path: Union[str, Path], search_path: Optional[Union[str, Path]] = None
) -> Union[Experiment, List[Experiment]]:
    """Loads all configuration and metrics outputs in an experiment directory.

    Parameters
    ----------
    exp_path: Union[str, Path]
        The directory to search for data. Directory must include the
        ".hydra/config.yaml" file.

    Returns
    ----------
    exps: Union[Experiment, List[Experiment]]

    """
    assert Path(exp_path).exists(), f"{exp_path} not found"

    # first find all .hydra files
    if search_path is None:
        search_path = ".hydra"
    cfg_files = sorted(Path(exp_path).absolute().glob(f"**/{str(search_path)}"))

    # For each file load metrics data
    exps = []
    for path in cfg_files:
        # Save experiment configuration
        cfg_files = list(path.parent.glob("**/config.yaml"))
        cfg = None
        if len(cfg_files) == 1:
            cfg = load_from_yaml(cfg_files[0])

        # Load metrics files
        files = path.parent.glob("*.pt")
        metrics = dict()
        for f in files:
            name = f.name
            metrics[name[:-3]] = torch.load(f)

        # Load path to checkpoints
        ckpts = [str(ckpt.resolve()) for ckpt in path.parent.glob("**/*.ckpt")]

        # Append experiment to list
        exps.append(Experiment(str(path.parent.parent), cfg, ckpts, metrics))

    if len(exps) == 1:
        return exps[0]

    return exps



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

import torch
from pytorch_lightning import LightningDataModule, LightningModule
from torch.nn import functional as fnn
from torch.utils.data import DataLoader, Dataset


class RandomDataset(Dataset):
    def __init__(self, size: int, length: int):
        self.len = length
        self.data = torch.randn(length, size)

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len


class SimpleLightningModule(LightningModule):
    def __init__(self):
        super().__init__()
        self.layer = torch.nn.Linear(32, 2)

    def forward(self, x):
        return self.layer(x)

    def loss(self, batch, prediction):
        # An arbitrary loss to have a loss that updates the model weights during `Trainer.fit` calls
        return fnn.mse_loss(prediction, torch.ones_like(prediction))

    def training_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        self.log("Tensor Metric", torch.tensor(1.0))
        self.log("Tensor 1 Metric", torch.tensor([1.0]))
        self.log("Number Metric", 1.0)
        return {"loss": loss}

    def training_step_end(self, training_step_outputs):
        return training_step_outputs

    def training_epoch_end(self, outputs: dict) -> None:
        torch.stack([x["loss"] for x in outputs]).mean()

    def validation_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        self.log("Val Tensor Metric", torch.tensor(1.0))
        self.log("Val Tensor 1 Metric", torch.tensor([1.0]))
        self.log("Val Number Metric", 1.0)
        return {"x": loss}

    def validation_epoch_end(self, outputs: dict) -> None:
        torch.stack([x["x"] for x in outputs]).mean()

    def test_step(self, batch, batch_idx):
        output = self(batch)
        loss = self.loss(batch, output)
        self.log("Tensor Metric", torch.tensor(1.0))
        self.log("Tensor 1 Metric", torch.tensor([1.0]))
        self.log("Number Metric", 1.0)
        return {"y": loss}

    def test_epoch_end(self, outputs: dict) -> None:
        torch.stack([x["y"] for x in outputs]).mean()

    def configure_optimizers(self):
        optimizer = torch.optim.SGD(self.layer.parameters(), lr=0.1)
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1)
        return [optimizer], [lr_scheduler]

    def train_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def val_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def test_dataloader(self):
        return DataLoader(RandomDataset(32, 64))


class SimpleDataModule(LightningDataModule):
    def train_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def val_dataloader(self):
        return DataLoader(RandomDataset(32, 64))

    def test_dataloader(self):
        return DataLoader(RandomDataset(32, 64))



# Copyright 2022, MASSACHUSETTS INSTITUTE OF TECHNOLOGY
# Subject to FAR 52.227-11 – Patent Rights – Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union

from torchvision.datasets import folder


class ImageNet(folder.DatasetFolder):
    def __init__(
        self,
        root: Union[str, Path],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        loader: Callable[[str], Any] = folder.default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ):
        """
        Parameters
        ----------
        root : Union[str, Path]
            Root directory of dataset where
            ImageNet folders exist.

        transform :  Optional[Callable[[Image], Any]]
            A function/transform that takes in a PIL image
            and returns a transformed version. E.g., ``transforms.RandomCrop``

        target_transform : Optional[Callable]
            A function/transform that takes in a target
            and returns a transformed version.

        loader : Callable[[str], Any], optional (default=`folder.default_loader`)
            A function to load a sample given its path.

        is_valid_file : Optional[Callable[[str], bool]], optional (default=None)
            A function that takes path of a file
            and checks if the file is a valid file.
        """
        super().__init__(
            root,
            loader,
            folder.IMG_EXTENSIONS if is_valid_file is None else None,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
        )


def remap_classes(
    range_sets: Tuple[Set[int], ...], class_to_idx: Dict[str, int]
) -> Tuple[List[str], Dict[str, int]]:
    """Filters existing classes and maps them to new index based on a pre-defined mapping.

    Parameters
    ----------
    range_sets: Tuple[Set[int], ...]
        Each element of the tuple is a list of indices that map the original class to the new class.

    class_to_idx: Dict[str, int]
        Dictionary mapping class name to class index

    Returns
    -------
    Tuple[List[str], Dict[str, int]]
        List of classes and dictionary mapping each class to an index.
    """
    mapping = {}
    for class_name, idx in class_to_idx.items():
        for new_idx, range_set in enumerate(range_sets):
            if idx in range_set:
                mapping[class_name] = new_idx
                break

    filtered_classes = list(mapping.keys())
    filtered_classes.sort()
    return filtered_classes, mapping


class _ReducedImageNet(folder.DatasetFolder):
    _range_sets: Tuple[Set[int], ...]

    def __init__(
        self,
        root: Union[str, Path],
        transform: Optional[Callable] = None,
        target_transform: Optional[Callable] = None,
        loader: Callable[[str], Any] = folder.default_loader,
        is_valid_file: Optional[Callable[[str], bool]] = None,
    ):
        super().__init__(
            root=root,
            loader=loader,
            extensions=folder.IMG_EXTENSIONS if is_valid_file is None else None,
            transform=transform,
            target_transform=target_transform,
            is_valid_file=is_valid_file,
        )

    def find_classes(self, directory: str) -> Tuple[List[str], Dict[str, int]]:
        classes, class_to_idx = super().find_classes(directory)
        classes, class_to_idx = remap_classes(self._range_sets, class_to_idx)
        return classes, class_to_idx


class RestrictedImageNet(_ReducedImageNet):
    """Restricted ImageNet, see https://arxiv.org/abs/1906.00945


    **WARNING**: This dataset is imbalanced


    Dog:
        ImageNet Class Indices: 151-268
    Cat:
        ImageNet Class Indices: 281-285
    Frog:
        ImageNet Class Indices: 30-32
    Turtle:
        ImageNet Class Indices: 33-37
    Bird:
        ImageNet Class Indices: 80-100
    Monkey:
        ImageNet Class Indices: 365-382
    Fish:
        ImageNet Class Indices: 389-397
    Crab:
        ImageNet Class Indices: 118-121
    Insect:
        ImageNet Class Indices: 300-319

    """

    _range_sets = tuple(
        set(range(s, e + 1))
        for s, e in [
            (151, 268),
            (281, 285),
            (30, 32),
            (33, 37),
            (80, 100),
            (365, 382),
            (389, 397),
            (118, 121),
            (300, 319),
        ]
    )


class ImageNetM10(_ReducedImageNet):
    """ImageNet-M10, see https://arxiv.org/pdf/2112.15329.pdf

    ImageNet-M10 consists of ten super-classes, each corresponding a WordNet ID in the hierarchy.
    Different super-classes contain varying number of ImageNet classes, so the dataset is balanced by
    choosing six classes within each super-class.

    Dog:
        WordNet ID: n02084071
        ImageNet Class Indices: 151 to 156
    Bird:
        WordNet ID: n01503061
        ImageNet Class Indices: 7 to 12
    Insect:
        WordNet ID: n02159955
        ImageNet Class Indices: 300 to 305
    Monkey:
        WordNet ID: n02484322
        ImageNet Class Indices: 370 to 375
    Car:
        WordNet ID: n02958343
        ImageNet Class Indices: 407, 436, 468, 511, 609, 627
    Feline:
        WordNet ID: n02120997
        ImageNet Class Indices: 286 to 291
    Truck:
        WordNet ID: n04490091
        ImageNet Class Indices: 555, 569, 675, 717, 734, 864
    Fruit:
        WordNet ID: n13134947
        ImageNet Class Indices: 948, 984, 987 to 990
    Fungus:
        WordNet ID: n12992868
        ImageNet Class Indices: 991 to 996
    Boat:
        WordNet ID: n02858304
        ImageNet Class Indices: 472, 554, 576, 625, 814, 914

    """

    _range_sets = (
        set(range(151, 156 + 1)),
        set(range(7, 12 + 1)),
        set(range(300, 305 + 1)),
        set(range(370, 375 + 1)),
        {407, 436, 468, 511, 609, 627},
        set(range(286, 291 + 1)),
        {555, 569, 675, 717, 734, 864},
        {948, 984, 987, 988, 999, 990},
        set(range(991, 996 + 1)),
        {472, 554, 576, 625, 814, 914},
    )
